import grpc # type: ignore

from grpc import (  # type: ignore
    UnaryUnaryClientInterceptor,
    UnaryStreamClientInterceptor,
    StreamUnaryClientInterceptor,
    StreamStreamClientInterceptor
)

from github.com.metaprov.modelaapi.services.grpcinferenceservice.v1 import grpcinferenceservice_pb2
from github.com.metaprov.modelaapi.services.grpcinferenceservice.v1 import grpcinferenceservice_pb2_grpc

from typing import Dict, Optional, Union, List

SERVER_ADDRESS = '0.0.0.0'
PORT = 8080


from dataclasses import dataclass

########################################################
# Server data
########################################################

@dataclass
class ServerInfo:
    name: str

@dataclass
class Metric:
    name: str
    value: float


@dataclass
class ModelInfo:
    name: str
    version: str
    product: str
    trainingMetrics: List[Metric]
    trainingScore: float
    status: str
    canary: bool
    shadow: bool
    traffic: float
    filter:str
    rank:int
    logPath: str
    p95: float
    p99: float
    lastPrediction: int
    dailyPredictionAvg: float

@dataclass
class ColumnInfo:
    name: str
    type: str
    datasetMin: float
    datasetMean: float
    datasetStdDev: float
    datasetMax: float


@dataclass
class SchemaInfo:
    columns: List[ColumnInfo]

@dataclass
class PredictorInfo:
    name: str
    task: str
    models: List[ModelInfo]
    schema: SchemaInfo

@dataclass
class OnePrediction:
    payload:str

@dataclass
class Predictions:
    payload:str # result as json
    items: List[OnePrediction]


#######################################################
# Model data
#######################################################


class ModelaGrpcClient(object):
    def __init__(self,
                 host,
                 port=3000,
                 interceptors: Optional[List[Union[
                     UnaryUnaryClientInterceptor,
                     UnaryStreamClientInterceptor,
                     StreamUnaryClientInterceptor,
                     StreamStreamClientInterceptor]]] = None
                 ):
        """Initializer.
           Creates a gRPC channel for connecting to the server.
           Adds the channel to the generated client stub.

        Arguments:
            None.

        Returns:
            None.
        """
        self._channel = grpc.insecure_channel(f'{host}:{port}')

        if interceptors:
            self._channel = grpc.intercept_channel(  # type: ignore
                self._channel, *interceptors)

        self._stub = grpcinferenceservice_pb2_grpc.GRPCInferenceServiceStub(self._channel)

    def close(self):
        """Closes Dapr runtime gRPC channel."""
        if self._channel:
            self._channel.close()

    def __del__(self):
        self.close()

    def __enter__(self) -> 'ModelaGrpcClient':
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self.close()

    def is_alive(self) -> bool:
        """Check if the service is alive

        Arguments:
           None

        Returns:
           bool; True if the predictor is alive
        """
        request = grpcinferenceservice_pb2.ServiceLiveRequest()
        try:
            response = self._stub.ServiceLive(request)
            return response.live
        except grpc.RpcError as err:
            print(err.details())  # pylint: disable=no-member
            print('{}, {}'.format(err.code().name, err.code().value))  # pylint: disable=no-member
            return False

    def is_ready(self) -> bool:
        """Check if the service is ready

        Arguments:
             None

        Returns:
            bool; True if the predictor is alive
        """
        request = grpcinferenceservice_pb2.ServiceReadyRequest()
        try:
            response = self._stub.ServerReady(request)
            return response.ready
        except grpc.RpcError as err:
            print(err.details())  # pylint: disable=no-member
            print('{}, {}'.format(err.code().name, err.code().value))  # pylint: disable=no-member
            return False

    def model_ready(self,name:str,version:str) -> bool:
        """Check if a specific model is ready

        Arguments:
             None

        Returns:
            bool; True if the predictor is alive
        """

        request = grpcinferenceservice_pb2.ModelReadyRequest()
        request.name    = name
        request.version = version
        try:
            response = self._stub.ModelReady(request)
            return response.ready
        except grpc.RpcError as err:
            print(err.details())  # pylint: disable=no-member
            print('{}, {}'.format(err.code().name, err.code().value))  # pylint: disable=no-member
            return False

    def server_metadata(self) -> ServerInfo:
        """Answer the server metadata info

        Arguments:
             None

        Returns:
            bool; True if the predictor is alive
        """

        request = grpcinferenceservice_pb2.ServerMetadataRequest()
        try:
            response = self._stub.ServerMetadata(request)
            return response.ready
        except grpc.RpcError as err:
            print(err.details())  # pylint: disable=no-member
            print('{}, {}'.format(err.code().name, err.code().value))  # pylint: disable=no-member


    def model_metadata(self,name:str,version:str) -> ModelInfo:
        request = grpcinferenceservice_pb2.ModelMetadataRequest()
        request.name = name
        request.version = version
        try:
            response = self._stub.ModelMetadata(request)
            return response.ready
        except grpc.RpcError as err:
            print(err.details())  # pylint: disable=no-member
            print('{}, {}'.format(err.code().name, err.code().value))  # pylint: disable=no-member
            return False


    def get_predictor(self) -> PredictorInfo:
        request = grpcinferenceservice_pb2.GetPredictorRequest()
        try:
            response = self._stub.GetPredictor(request)
            return response.item
        except grpc.RpcError as err:
            print(err.details())  # pylint: disable=no-member
            print('{}, {}'.format(err.code().name, err.code().value))  # pylint: disable=no-member
            raise err

    def get_model(self,predictor,name) -> ModelInfo:
        """Answer the model information

        Arguments:
             name - the name of the model

        Returns:
            bool; True if the predictor is alive
        """

        request = grpcinferenceservice_pb2.GetModelRequest()
        request.predictorName = predictor
        request.name          = name
        try:
            response = self._stub.GetModel(request)
            return response.item
        except grpc.RpcError as err:
            print(err.details())  # pylint: disable=no-member
            print('{}, {}'.format(err.code().name, err.code().value))  # pylint: disable=no-member
            raise err


    def predict(self,
                payload:str,
                name="",
                validate=False,
                explain=False,
                format="json",
                labeled=False,
                metrics=None) -> Predictions:
        request = grpcinferenceservice_pb2.PredictRequest()
        request.name      = name
        request.validate  = validate
        request.explain   = explain
        request.format    = format
        request.payload   = payload
        request.labeled   = labeled
        request.metrics.extend(metrics)

        try:
            response = self._stub.Predict(request)
            return response.items
        except grpc.RpcError as err:
            print(err.details())  # pylint: disable=no-member
            print('{}, {}'.format(err.code().name, err.code().value))  # pylint: disable=no-member
            raise err


import _collections_abc
import math


# Source: https://stackoverflow.com/questions/5194057/better-way-to-convert-file-sizes-in-python
def convert_size(size_bytes):
   if size_bytes == 0:
       return "0B"
   size_name = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
   i = int(math.floor(math.log(size_bytes, 1024)))
   p = math.pow(1024, i)
   s = round(size_bytes / p, 2)
   return "%s %s" % (s, size_name[i])

class TrackedList(_collections_abc.MutableSequence):
    """ TrackedList is a wrapper for a list that will propagate changes to a Configuration. """

    def __init__(self, data, parent, attribute):
        self.data = []
        self._parent = parent
        self._parent_attr = attribute
        if data is not None:
            if type(data) == type(self.data):
                self.data[:] = data
            elif isinstance(data, TrackedList):
                self.data[:] = data.data[:]
            else:
                print(type(data))
                self.data = list(data)

        self.propagate()

    def __repr__(self):
        return repr(self.data)

    def __lt__(self, other):
        return self.data < self.__cast(other)

    def __le__(self, other):
        return self.data <= self.__cast(other)

    def __eq__(self, other):
        return self.data == self.__cast(other)

    def __gt__(self, other):
        return self.data > self.__cast(other)

    def __ge__(self, other):
        return self.data >= self.__cast(other)

    def __cast(self, other):
        return other.data if isinstance(other, TrackedList) else other

    def __contains__(self, item):
        return item in self.data

    def __len__(self):
        return len(self.data)

    def __getitem__(self, i):
        if isinstance(i, slice):
            return self.__class__(self.data[i])
        else:
            return self.data[i]

    def __setitem__(self, i, item):
        self.data[i] = item
        self.propagate()


    def __delitem__(self, i):
        del self.data[i]
        self.propagate()


    def __add__(self, other):
        if isinstance(other, TrackedList):
            return self.__class__(self.data + other.data, self._parent, self._parent_attr)
        elif isinstance(other, type(self.data)):
            return self.__class__(self.data + other, self._parent, self._parent_attr)
        return self.__class__(self.data + list(other), self._parent, self._parent_attr)

    def __radd__(self, other):
        if isinstance(other, TrackedList):
            return self.__class__(other.data + self.data, self._parent, self._parent_attr)
        elif isinstance(other, type(self.data), self._parent, self._parent_attr):
            return self.__class__(other + self.data, self._parent, self._parent_attr)
        return self.__class__(list(other) + self.data, self._parent, self._parent_attr)

    def __iadd__(self, other):
        if isinstance(other, TrackedList):
            self.data += other.data
        elif isinstance(other, type(self.data)):
            self.data += other
        else:
            self.data += list(other)
        return self

    def __mul__(self, n):
        return self.__class__(self.data * n, self._parent, self._parent_attr)

    __rmul__ = __mul__

    def __imul__(self, n):
        self.data *= n
        return self

    def __copy__(self):
        inst = self.__class__.__new__(self.__class__)
        inst.__dict__.update(self.__dict__)
        # Create a copy and avoid triggering descriptors
        inst.__dict__["data"] = self.__dict__["data"][:]
        return inst

    def propagate(self):
        if self._parent:
            self._parent.propagate_to_parent(self._parent_attr, self.data)

    def append(self, item):
        self.data.append(item)
        self.propagate()

    def insert(self, i, item):
        self.data.insert(i, item)
        self.propagate()


    def pop(self, i=-1):
        p = self.data.pop(i)
        self.propagate()
        return p

    def remove(self, item):
        self.data.remove(item)
        self.propagate()


    def clear(self):
        self.data.clear()
        self.propagate()

    def copy(self):
        return self.__class__(self, self._parent, self._parent_attr)

    def count(self, item):
        return self.data.count(item)

    def index(self, item, *args):
        return self.data.index(item, *args)

    def reverse(self):
        self.data.reverse()

    def sort(self, /, *args, **kwds):
        self.data.sort(*args, **kwds)

    def extend(self, other):
        if isinstance(other, TrackedList):
            self.data.extend(other.data)
        else:
            self.data.extend(other)



from dataclasses import field
from typing import List
from github.com.metaprov.modelaapi.pkg.apis.catalog.v1alpha1.generated_pb2 import Measurement
from github.com.metaprov.modelaapi.services.common.v1.common_pb2 import ModelProfile
from modela.Configuration import Configuration, ImmutableConfiguration, datamodel
from modela.common import PriorityLevel, Time, StatusError, ConditionStatus, ObjectReference, Freq, Plot
from modela.data.common import DataType
from modela.data.models import DataLocation, GovernanceSpec, CompilerSettings, Correlation, DataSourceSpec
from modela.inference.common import AccessType
from modela.infra.models import Workload, OutputLogs, NotificationSettings
from modela.training.common import *
import github.com.metaprov.modelaapi.pkg.apis.catalog.v1alpha1.generated_pb2 as catalog_pb
import github.com.metaprov.modelaapi.pkg.apis.training.v1alpha1.generated_pb2 as training_pb
import github.com.metaprov.modelaapi.services.common.v1.common_pb2 as common_pb


@datamodel(proto=training_pb.ModelValidation)
class ModelValidation(Configuration):
    """ ModelValidation defines a single validation to be run against a model """
    Type: ModelValidationType = None
    """ The type of model validation """
    PrevModel: str = ''
    """ PrevModel specifies a previous model to compare against """
    DatasetName: str = ''
    """ The name of a labeled dataset used to test the model, when measuring a performance metric """
    DriftFreq: Freq = None
    DriftInterval: int = 0
    Column: str = ''
    Metric: Metric = None
    Min: float = 0
    Max: float = 0
    MinPercent: float = 0
    MaxPercent: float = 0
    Agg: Aggregate = None
    """ Agg specifies the type of aggregate when measuring aggregate performance (e.g. median, average) """


@datamodel(proto=training_pb.ModelValidationResult)
class ModelValidationResult(Configuration):
    Type: str = ''
    DatasetName: str = ''
    """ DatasetName is the name of the dataset used to perform the validation """
    ModelName: str = ''
    """ ModelName is the name of the model under test """
    Column: str = ''
    """ Column is the name of the feature. """
    Error: str = ''
    """ Error is a string. """
    Metric: Metric = None
    """ Metric is the measurement """
    ActualValue: float = 0
    """ Actual value is the actual value """
    Passed: bool = False
    """ Passed indicate if the result passed. """
    At: Time = None
    """ At is the time the validation was performed. """
    DurationInSec: int = 0
    """ The amount of time it took to compute this result """


@datamodel(proto=catalog_pb.Measurement)
class Measurement(ImmutableConfiguration):
    """ Measurement is a value for a specific metric """
    Metric: Metric = Metric.Null
    """ The metric type name (e.g. F1 / Accuracy) """
    Value: float = 0
    """ The value of the metric """


@datamodel(proto=training_pb.SegmentSpec)
class Segment(Configuration):
    ColumnName: str = ''
    Op: Operation = Operation.EQ
    """ The segment operation """
    Value: str = ''
    """ The value operation """


@datamodel(proto=training_pb.HyperParameterValue)
class HyperParameterValue(ImmutableConfiguration):
    """ HyperParameterValue represent an individual hyper-parameter of a machine earning algorithm """
    Name: str = ''
    """ The name of the hyper-parameter """
    Value: str = ''
    """ The value of the hyper-parameter """


@datamodel(proto=training_pb.ClassicalEstimatorSpec)
class ClassicalEstimatorSpec(ImmutableConfiguration):
    """ ClassicalEstimatorSpec is the specification for an algorithm and the actual value fof the hyper parameters """
    AlgorithmName: str = ''
    """ AlgorithmName is a reference to the algorithm in the catalog """
    Parameters: List[HyperParameterValue] = field(default_factory=lambda : [])
    """ Parameters is a list of the algorithm hyper parameters """


@datamodel(proto=training_pb.ChatbotEstimatorSpec)
class ChatbotEstimatorSpec(ImmutableConfiguration):
    """ ChatbotEstimatorSpec specifies the configuration for a chatbot model """
    Base: str = ''
    """ The name of the base model """


@datamodel(proto=training_pb.NLPEstimatorSpec)
class NLPEstimatorSpec(ImmutableConfiguration):
    """ NLPEstimatorSpec specifies the configuration for an NLP model """
    Base: str = ''
    """ The name of the base model """


@datamodel(proto=training_pb.FeatureImportance)
class FeatureImportance(ImmutableConfiguration):
    """ FeatureImportance records the computed importance of a single feature """
    Feature: str = ''
    """ The name of the feature """
    Importance: float = 0
    """ The importance of the feature """


@datamodel(proto=training_pb.SuccessiveHalvingSpec)
class SuccessiveHalving(ImmutableConfiguration):
    """ SuccessiveHalvingSpec records the position of a single model in a successive halving search """
    Budget: int = 0
    """
    The budget to be used in a multi-fidelity search, for example the number of epochs
    or the percentage of data to train the model with
    """
    Bracket: int = 0
    """ The bracket of the model """
    Rung: int = 0
    """ The rung of the model """
    ConfID: int = 0
    """ The configuration ID allocated to the model """
    Modality: ModalityType = None
    """
    The type of modality, based on the type of model
    For classical models, it should be based on data percentage
    For deep models, it should be based on epochs
    """


@datamodel(proto=training_pb.DataSplitSpec)
class DataSplit(Configuration):
    """ DataSplitSpec specifies the configuration to split a dataset into training and testing datasets """
    Method: DataSplitMethod = DataSplitMethod.Auto
    """ The type of split method """
    Train: int = 80
    """ The number percentage (0 through 100) of rows that will be allocated to the training dataset """
    Validation: int = 0
    """
    The number percentage (0 through 100) of rows that will be allocated to the validation dataset.
    If Validation is set to 0 the model will default to use cross-validation
    """
    Test: int = 20
    """ The number percentage (0 through 100) of rows that will be allocated to the training dataset """
    SplitColumn: str = ''
    """
    The name of the column containing a binary value that indicates if the row should be split.
    The split type must use split-column in order for SplitColumn to have an effect
    """
    Segments: List[Segment] = field(default_factory=lambda : [])
    """ The collection of segments """
    TrainDataset: str = ''
    """ The name of the Dataset resource which will be used as the training dataset """
    TestDataset: str = ''
    """
    The name of the Dataset resource which will be used as the testing dataset, applicable
    if the split type uses test-dataset. If enabled, the training dataset will not be split and used as-is
    """


@datamodel(proto=training_pb.TrainingSpec)
class Training(Configuration):
    """ TrainingSpec specifies the configuration of a model training workload """
    Priority: PriorityLevel = PriorityLevel.Medium
    """ The priority of the Kubernetes Job created by the Model (medium, by default) """
    Cvtype: CvType = CvType.CVTypeKFold
    """ The type of cross-validation to use, in the case that a validation dataset is not enabled """
    CV: bool = True
    """ Indicates if cross-validation should be used to score models """
    Folds: int = 5
    """ The number of folds to use during cross-validation """
    Split: DataSplit = DataSplit()
    """ Split specifies the configuration to generate training, testing, and validation datasets """
    EvalMetrics: List[Metric] = field(default_factory=lambda : [])
    """ EvalMetrics specifies the collection of metrics that will be evaluated after model training is complete """
    EarlyStop: bool = False
    """ Indicates if the parent Study should stop sampling new models if there is no improvement in score """
    CheckpointInterval: int = 10
    """
    The interval, in minutes, at which a snapshot of a partially trained model will be saved.
    Applicable to models with long training times for resiliency if training is suddenly stopped
    """
    Sh: SuccessiveHalving = None
    """ SuccessiveHalving specifies the configuration for a Study to execute a model search using successive halving """
    Seed: float = 42
    """ The random state used for the model's estimator, if applicable (for example, in RandomForestClassifier) """
    Resources: Workload = Workload('general-large')
    """ Resources specifies the resource requirements allocated to the model training workload """
    Gpu: bool = False
    """ Indicates if a GPU will be allocated for model training """
    Distributed: bool = False
    """ Indicates if model training will be distributed across multiple nodes (currently unimplemented) """
    NodeCount: int = 1
    """ The number of nodes to use, in the case of distributed training """
    SamplePct: int = 100
    """ The number percentage (0 through 100) of rows to be used during training """
    LabRef: ObjectReference = ObjectReference('default-tenant', 'default-lab')
    """ The reference to the Lab under which the model training Job will be created """


@datamodel(proto=training_pb.ServingSpec)
class ServingSpec(ImmutableConfiguration):
    """ ServingSpec specifies the requirements to serve a model """
    Resources: Workload = None


@datamodel(proto=training_pb.TextPipelineSpec)
class TextPipelineSpec(ImmutableConfiguration):
    """ TextPipelineSpec represents a single pipeline for transforming text data """
    Encoder: TextEncoding = TextEncoding.Auto
    """ The text encoder (e.g. TFIDF) """
    Tokenizer: str = ''
    """ The text tokenizer character """
    Stopwords: bool = True
    """ Indicates if the pipeline will add stop word handling """
    Pos: bool = True
    """ Indicates if the pipeline will add part-of-speech handling """
    Lemma: bool = True
    """ Indicates if the pipeline will implement lemmatization """
    Stem: bool = True
    """ Indicates if the pipeline will implement word stemming """
    Embedding: str = ''
    """ Indicates if the pipeline will generate a word embedding """
    Svd: bool = True
    """ Indicates if the pipeline will add singular value decomposition """
    MaxSvdComponents: int = 0
    """ The maximum number of SVD components to use, if applicable """


@datamodel(proto=training_pb.ImagePipelineSpec)
class ImagePipelineSpec(ImmutableConfiguration):
    """ ImagePipelineSpec represents a single pipeline for preprocessing image data """
    Featurizer: str = None
    """ The date time imputer. """


@datamodel(proto=training_pb.VideoPipelineSpec)
class VideoPipelineSpec(ImmutableConfiguration):
    """ VideoPipelineSpec represents a single pipeline for preprocessing video data """
    Featurizer: str = None
    """ The date time imputer. """


@datamodel(proto=training_pb.AudioPipelineSpec)
class AudioPipelineSpec(ImmutableConfiguration):
    """ AudioPipelineSpec represents a single pipeline for preprocessing audio data """
    Featurizer: str = None
    """ The date time imputer. """


@datamodel(proto=training_pb.ResourceConsumption)
class ResourceConsumption(ImmutableConfiguration):
    """ ResourceConsumption represents the total resources consumed by a workload """
    Cpu: float = 0
    Mem: float = 0
    Gpu: float = 0


@datamodel(proto=training_pb.DataHashes)
class DataHashes(ImmutableConfiguration):
    """ DataHashes contains the hashes for datasets used during model training """
    TrainHash: str = ''
    TestingHash: str = ''
    ValidationHash: str = ''


@datamodel(proto=training_pb.GeneratedColumnSpec)
class GeneratedColumnSpec(ImmutableConfiguration):
    """ GeneratedColumnSpec describes a column to be generated and applied to a dataset """
    Name: str = ''
    """ The name of the generated column """
    Datatype: DataType = DataType.Text
    """ The resulting data type """
    First: str = ''
    """ The name of the first original column """
    Second: str = ''
    """ The name of the second original column, if the expression is binary operator """
    Original: str = ''
    """ The expression to apply in order to generate the new column """


@datamodel(proto=training_pb.FeatureSelectionSpec)
class FeatureSelection(Configuration):
    """ FeatureSelectionSpec specifies the configuration to run feature selection on a dataset """
    Enabled: bool = False
    """ Indicates if feature selection is enabled """
    SamplePct: int = 100
    """ The number percentage (0 through 100) of the dataset to sample """
    Embedding: bool = False
    """ Indicates if embedded methods will be tested as part of the candidate algorithms (e.g. tree-based selection) """
    Filter: bool = False
    """ Indicates if filter methods will be tested as part of the candidate algorithms (e.g. chi-square or anova tests) """
    Wrapper: bool = False
    """ Indicates if wrapper methods will be tested as part of the candidate algorithms """
    Pipeline: List[FeatureSelectionType] = field(default_factory=lambda : [])
    """ The collection of feature selection methods that will be applied in order to the dataset """
    VarianceThresholdPct: int = 5
    """ The threshold as a percentage to remove low variance features """
    CorrThreshold: int = 95
    """ The threshold to remove features with high correlations """
    TopN: int = 0
    """ The number of features that will be selected based on importance. If TopN is 0, all features will be selected """
    CumulativeImportancePercent: int = 95
    """ The cumulative importance threshold of features to be included """
    Reserved: List[str] = field(default_factory=lambda : [])
    """ List of features that are reserved and will always be included in the final feature selection """


@datamodel(proto=training_pb.FeaturePair)
class FeaturePair(ImmutableConfiguration):
    X: str = ''
    Y: str = ''


@datamodel(proto=training_pb.InterpretabilitySpec)
class Interpretability(Configuration):
    """ InterpretabilitySpec specifies the configuration to generate interpretability data and diagrams """
    Ice: bool = True
    """ Indicates if ICE (individual condition expectation) plots will be generated """
    Icepairs: List[FeaturePair] = field(default_factory=lambda : [])
    """ The collection of feature pairs to generate ICE scatter diagrams for each """
    Lime: bool = False
    """ Indicates if LIME (local interpretable model-agnostic explanations) diagrams will be generated """
    Shap: ShapType = ShapType.Auto
    """
    The type of SHAP values to be generated. Linear and tree values are the
    only recommended types due to the high compute times of other methods
    """
    Shappairs: List[FeaturePair] = field(default_factory=lambda : [])
    """ The collection of feature pairs to generate SHAP scatter diagrams for each """
    Counterfactual: bool = False
    """ Indicates if counter-factual diagrams will be generated """
    Anchor: bool = False
    """ Indicates if anchor explanation diagrams will be generated """


ImputationType = Imputation
DiscretisationType = Discretisation
OutlierHandlingType = OutlierHandling
VariableTransformationType = VariableTransformation
DatetimeTransformationType = DatetimeTransformation
ScalingType = Scaling


@datamodel(proto=training_pb.FeatureEngineeringPipeline)
class FeatureEngineeringPipeline(ImmutableConfiguration):
    """ FeatureEngineeringPipeline represent a single pipeline to transform a dataset """
    Name: str = ''
    """ The name of the pipeline """
    Datatype: DataType = DataType.Text
    """ The type of data which the pipeline applies to """
    Columns: List[str] = field(default_factory=lambda : [])
    """
    The collection of columns which the pipeline applies to. Each column in the dataset with the
    data type of the pipeline should be added to the collection of columns
    """
    Imputation: ImputationType = ImputationType.AutoImputer
    """ The imputation method to use, which fills in missing values within the columns """
    Encoding: CategoricalEncoding = CategoricalEncoding.NoneEncoding
    """ The encoding method to use for categorical data types """
    Scaling: ScalingType = ScalingType.NoScaling
    """ The scaling method to use for numerical data types """
    Discretisation: DiscretisationType = DiscretisationType.NoDiscretisation
    """ The discretisation method, which converts numerical data types to discrete variables """
    VariableTransformation: VariableTransformationType = VariableTransformationType.NoneTransform
    """ The transformation method to use for numerical data types """
    OutlierHandling: OutlierHandlingType = OutlierHandlingType.AutoOutlier
    """
    The method to use when handling outliers
    Apply only to numeric datatypes.
    """
    DatetimeTransformation: DatetimeTransformationType = DatetimeTransformationType.NoneDatetime
    """ The method to use when handling the date-time data type """
    Text: TextPipelineSpec = None
    """ Text specifies the pipeline to handle raw text """
    Image: ImagePipelineSpec = None
    """ Image specifies the pipeline to handle image data (currently unsupported) """
    Audio: AudioPipelineSpec = None
    """ Audio specifies the pipeline to handle audio data (currently unsupported) """
    Video: VideoPipelineSpec = None
    """ Video specifies the pipeline to handle video data (currently unsupported) """
    Generated: List[GeneratedColumnSpec] = field(default_factory=lambda : [])
    """ Generated specifies a collection of columns to be generated """
    Custom: List[GeneratedColumnSpec] = field(default_factory=lambda : [])
    """ Custom specifies a collection of columns to be generated. Custom columns are specified by end-users """
    Drop: bool = False
    """ Indicates if all of all the columns specified by the Columns field should be dropped """
    Passthrough: bool = False
    """ Indicates if the pipeline should not be applied and the columns should remain unchanged """


@datamodel(proto=training_pb.FeatureEngineeringSpec)
class FeatureEngineeringSpec(ImmutableConfiguration):
    """ FeatureEngineeringSpec specifies the feature engineering and preprocessing to be performed on a dataset """
    Pipelines: List[FeatureEngineeringPipeline] = field(default_factory=lambda : [])
    """
    Pipelines contains the collection of feature engineering pipelines that
    will be applied to a dataset prior to model training
    """
    Imbalance: ImbalanceHandling = ImbalanceHandling.ImbalanceAuto
    """ The method which will be used to handle an imbalanced dataset """
    Selection: FeatureSelection = FeatureSelection()
    """
    FeatureSelection specifies the configuration to perform
    feature selection on a dataset prior to model training
    """


@datamodel(proto=training_pb.EnsembleSpec)
class EnsembleSpec(ImmutableConfiguration):
    """ EnsembleSpec specifies the parameters of an ensemble model """
    Models: List[str] = field(default_factory=lambda : [])
    """ The collection of models (by their name) to be used as estimators in the ensemble """
    Estimators: List[ClassicalEstimatorSpec] = field(default_factory=lambda : [])
    """ The collection of estimators to be used in the ensemble, derived from Models """
    Base: ClassicalEstimatorSpec = None
    """ The base estimator """
    Type: EnsembleType = None
    """ The ensembling method """


TrainingSpec = Training
InterpretabilitySpec = Interpretability


@datamodel(proto=training_pb.ModelSpec)
class ModelSpec(ImmutableConfiguration):
    """ ModelSpec defines the desired state of the Model resource """
    Owner: str = 'no-one'
    """ The owner of the Study that created the Model """
    VersionName: str = 'v0.0.1'
    """
    The name of the DataProductVersion which describes the version of the resource
    that exists in the same DataProduct namespace as the resource, derived from the parent Study
    """
    ModelVersion: str = ''
    """ The user-assigned version of the Model, derived from the parent Study """
    StudyName: str = ''
    """ The name of the Study which created the Model. If empty, the Model will be trained as a stand-alone model """
    DatasetName: str = ''
    """ The name of the Dataset resource which the Model is being trained with """
    Task: TaskType = None
    """ The machine learning task type of the Model (i.e. regression, classification), derived from the parent Study """
    Objective: Metric = None
    """ The objective metric that will be tested and used to evaluate the performance of the model against others """
    FeatureEngineering: FeatureEngineeringSpec = None
    """
    FeatureEngineering specifies the preprocessing pipelines that will be applied to the model prior to training.
    By default, feature engineering is generated automatically by sampling different pipelines in competition
    """
    Estimator: ClassicalEstimatorSpec = None
    """ Estimator specifies the machine learning algorithm and hyper-parameters of the Model """
    NlpEstimator: NLPEstimatorSpec = None
    """ NLPEstimator specifies the estimator for a deep NLP model (currently unimplemented) """
    Ensemble: EnsembleSpec = None
    """ Ensemble specifies the configuration to create an ensemble model """
    Training: TrainingSpec = None
    """ TrainingSpec specifies the configuration to prepare and train a model """
    Serving: ServingSpec = None
    """
    ServingSpec defines the resource requirements that will be applied to a Predictor
    that will be created if the model is set to be released
    """
    Tested: bool = False
    """
    Tested indicates if a workload will be instantiated to test the model. The Study resource controller
    will automatically set this field if the Model was found to be the highest-performing
    """
    Aborted: bool = False
    """ Aborted indicates if any workloads associated with the Model should stop execution """
    Packaged: bool = False
    """ Packaged indicates if the Model should be packaged into tarbell """
    Published: bool = False
    """ Published indicates that the Model should be baked into a Docker image """
    Pushed: bool = False
    """ Pushed indicates that the Model image should be pushed to a Docker image registry """
    Reported: bool = False
    """ Reported indicates that a Report will be generated for the Model """
    Paused: bool = False
    """ Paused indicates that the execution of new workloads associated with the Model should be paused """
    Profiled: bool = False
    """ Profiled indicates that the Model will be profiled """
    Archived: bool = False
    """ Archived indicates that the Model should be archived in long-term storage """
    Forecasted: bool = False
    """ Forecasted indicates that the Model should perform a forecast """
    Released: bool = False
    """ Released indicates that the Model will be deployed within Predictor """
    Benchmarked: bool = False
    Explained: bool = False
    """ Explained indicates if a workload to compute SHAP values/diagrams should be executed """
    Baseline: bool = False
    """ Baseline indicates if the Model was produced by the baseline phase of a Study """
    Flagged: bool = False
    """ Indicates if the model is flagged """
    Location: DataLocation = None
    """ The data location where artifacts (metadata, reports, and estimators) generated by the Model will be stored """
    Compilation: CompilerSettings = None
    """ Compilation specifies the configuration to compile a model to a binary (currently unimplemented) """
    ActiveDeadlineSeconds: int = 600
    """ The deadline for any Jobs associated with the Model to be completed in seconds """
    EstimatorType: ModelType = ModelType.Classical
    """ ModelType is the type of model for this estimator """
    ModelClass: ModelClassType = None
    """ The model class, which is derived from the phase of the Study that the Model was created for """
    TrialID: int = 0
    """ The trial ID, which is incremented for each Model produced by the data plane """
    Governance: GovernanceSpec = None
    """ Governance specifies the model governance requirements (currently unimplemented) """
    Interpretability: InterpretabilitySpec = None
    """ Interpretability specifies the configuration to generate model interpretability visualizations """


@datamodel(proto=training_pb.InterpretabilityStatus)
class InterpretabilityStatus(ImmutableConfiguration):
    """ InterpretabilityStatus represents the state of the explanation phase of a Model """
    TrainingStartTime: Time = None
    """ StartTime represents the time when the model explanation phase started """
    TrainingEndTime: Time = None
    """ EndTime represents the time when the model explanation phase ended """
    ExplainerURI: str = ''
    """ The URI for the generated explanation data """
    TrainShapValuesURI: str = ''
    """ The URI for the train SHAP values """
    TestShapValuesURI: str = ''
    """ The URI for the test SHAP values """
    Importance: List[FeatureImportance] = field(default_factory=lambda : [])
    """ The collection of feature importances generated from the computed SHAP values """


@datamodel(proto=training_pb.ModelCondition)
class ModelCondition(ImmutableConfiguration):
    """ ModelCondition describes the state of a Model at a certain point """
    Type: ModelConditionType = ModelConditionType.ModelReady
    """ Type of Model condition """
    Status: ConditionStatus = ConditionStatus.ConditionUnknown
    """ Status of the condition, one of True, False, Unknown """
    LastTransitionTime: Time = None
    """ Last time the condition transitioned from one status to another """
    Reason: str = ''
    """ The reason for the condition's last transition """
    Message: str = ''
    """ A human-readable message indicating details about the transition """


@datamodel(proto=training_pb.ModelStatus)
class ModelStatus(ImmutableConfiguration):
    """ ModelStatus defines the observed state of a Model """
    StartTime: Time = None
    """ StartTime represents the time at which the execution of the Model started """
    TrainingStartTime: Time = None
    """ TrainStartTime represents the time at which the Model started training """
    TrainingEndTime: Time = None
    """ TrainCompletionTime represents the time at which the Model completed training """
    TestingStartTime: Time = None
    """ TestingStartTime represents the time at which the Model started testing """
    TestingEndTime: Time = None
    """ TestingEndTime represents the time at which the Model completed testing """
    EndTime: Time = None
    """ EndTime represents the time at which the Model was marked as ready, failed, or aborted """
    CvScore: float = 0
    """ The cross-validation score for the objective metric produced after training """
    TrainingScore: float = 0
    """ The score for the objective metric based on the training dataset """
    TestScore: float = 0
    """ The score for the objective metric based on the testing dataset """
    Cost: float = 0
    """ Cost is the cost of training the model in the case of a deep-learning model """
    Best: bool = False
    """ Best indicates if the Model was found to be the best model produced by a Study """
    Cv: List[Measurement] = field(default_factory=lambda : [])
    """
    CV contains the collection of measurements produced by cross-validation
    on the training dataset or validation on the validation dataset
    """
    Train: List[Measurement] = field(default_factory=lambda : [])
    """ Train contains the collection of measurements produced by validation on the training dataset """
    Test: List[Measurement] = field(default_factory=lambda : [])
    """ Train contains the collection of measurements produced by validation on the testing dataset """
    Phase: ModelPhase = ModelPhase.Pending
    """ The phase of the Model """
    ReportName: str = ''
    """ The name of the Report resource produced by the Model """
    ReportUri: str = ''
    """ The URI of the Report """
    ManifestUri: str = ''
    """ The URI of the Model manifest """
    WeightsUri: str = ''
    """ The URI of the model weights binary file """
    LabelEncoderUri: str = ''
    """ The URI of the label encoder binary file, if it exists """
    LogsUri: str = ''
    """ The URI of the logs file """
    ProfileUri: str = ''
    """ The URI of the model profile, which contains visualizations produced during the profiling phase """
    MisclassUri: str = ''
    """ The URI to the misclassification file produced during the testing phase """
    TarUri: str = ''
    """ The URI to the model tarbell file """
    AppUri: str = ''
    """ The URI to the model application file """
    ImageName: str = ''
    """ The name of the Docker image produced by the Model """
    Importance: List[FeatureImportance] = field(default_factory=lambda : [])
    """ The collection of features and their importance, sorted by the greatest importance first """
    ForecastUri: str = ''
    """ The URI of the model forecast """
    PythonVersion: str = ''
    """ The Python version of the data plane used during training """
    TrainDataset: DataLocation = None
    """ TrainDatasetLocation specifies the location of the training dataset """
    TestDataset: DataLocation = None
    """ TestDatasetLocation specifies the location of the testing dataset """
    ValidationDataset: DataLocation = None
    """ ValidationDataset specifies the location of the validation dataset """
    ObservedGeneration: int = 0
    """ ObservedGeneration is the last generation that was acted on """
    TrainingRows: int = 0
    """ The number of rows in the training dataset """
    TestingRows: int = 0
    """ The number of rows in the testing dataset """
    ValidationRows: int = 0
    """ The number of rows in the validation dataset, if applicable """
    FailureReason: StatusError = None
    """ In the case of failure, the Model resource controller will set this field with a failure reason """
    FailureMessage: str = ''
    """ In the case of failure, the Model resource controller will set this field with a failure message """
    Progress: int = 0
    """ The progress percentage of the Model, which is derived from the Model's current phase """
    SizeInBytes: int = 0
    """ The size of the model binary in bytes """
    Latency: float = 0
    """ The measured prediction latency """
    Url: str = ''
    """ The URL to the released model """
    PredictorName: str = ''
    """ The name of the Predictor in the case that the Model has been released and deployed """
    ReleasedAt: Time = None
    """ The time at which the Model was set to release """
    TarfileHash: str = ''
    """ Sha256 of the model tar file """
    ImageHash: str = ''
    """ Sha256 of the model image """
    TrainingDataHash: DataHashes = None
    """ TrainingDataHash specifies the hashes for datasets used by the Model """
    TrainingResources: ResourceConsumption = None
    """ TrainingResources details the resources that were consumed by the training workload """
    TestingResources: ResourceConsumption = None
    """ TestingResources details the resources that were consumed by the testing workload """
    TrainedBy: str = ''
    """ The Account which trained the model, derived from the parent Study """
    Team: str = ''
    """ The team of the Account which trained the model, derived from the parent Study """
    Endpoint: str = ''
    """ The endpoint of the Model, which is set after it is deployed to a Predictor """
    Logs: OutputLogs = None
    """ Logs specifies the location of logs produced by workloads associated with the Model """
    CorrelationsWithTarget: List[Correlation] = field(default_factory=lambda : [])
    """ The collection of correlations of the features of the training dataset and the target feature """
    TopCorrelations: List[Correlation] = field(default_factory=lambda : [])
    """ The top correlations between features of the training dataset """
    LastUpdated: Time = None
    """ The last time the object was updated """
    Interpretability: InterpretabilityStatus = None
    """ Interpretability contains results produced during the explaining phase of the Model """
    Conditions: List[ModelCondition] = field(default_factory=lambda : [])


@datamodel(proto=common_pb.ModelProfile)
class ModelProfile(ImmutableConfiguration):
    Name: str = ''
    Importance: List[float] = field(default_factory=lambda : [])
    Plots: List[Plot] = field(default_factory=lambda : [])


@datamodel(proto=training_pb.AlgorithmSearchSpaceSpec)
class AlgorithmSearchSpace(Configuration):
    """ AlgorithmSearchSpaceSpec defines the algorithms available to models produced by a Study """
    Allowlist: List[ClassicEstimator] = field(default_factory=lambda : [])
    """
    AllowList contains the collection of algorithms available to the parent Study.
    If AllowList is empty, all algorithms will be available for training
    """
    Filter: AlgorithmFilter = AlgorithmFilter.NoFilter


@datamodel(proto=training_pb.SuccessiveHalvingOptions)
class SuccessiveHalvingOptions(Configuration):
    """ SuccessiveHalvingOptions defines the parameters for a successive halving search """
    MaxBudget: int = 81
    """
    The maximum budget allocated to models during the successive halving search. For classical models, this
    represents the number percentage (0 through 100) of data that can be allocated to the model for training
    """
    EliminationRate: int = 3
    """
    The denominator for the fraction of models that will be promoted to the next round
    (i.e. an EliminationRate of 3 will only promote 1/3rd models to the next round)
    """
    Modality: ModalityType = ModalityType.Epochs
    """
    The type of modality, based on the type of model
    For classical models, it should be based on data percentage
    For deep models, it should be based on epochs
    """


@datamodel(proto=training_pb.PrunerSpec)
class PrunerSettings(Configuration):
    Type: Pruner = Pruner.MedianPruner
    StartupTrials: int = 5
    """ Pruning is disabled until the given number of trials finish in the same study. """
    WarmupTrials: int = 0
    """ Pruning is disabled until the trial exceeds the given number of step """
    MinTrials: int = 1
    """ Minimum number of reported trials. """
    IntervalTrials: int = 1
    """ Interval in number of steps between the pruning checks """
    Percentile: int = 25
    """ Keep specific percent of trials. Used only with percentile pruner """
    Lower: int = 0
    """ A minimum value which determines whether pruner prunes or not. Used only for threshold pruner """
    Upper: int = 0
    """ A maximum value which determines whether pruner prunes or not. Used only for threshold pruner """
    ShOptions: SuccessiveHalvingOptions = SuccessiveHalvingOptions()
    """
    SHOptions is the desired options for successive halving search.
    All other models are saved into an archive.
    """


SamplerType = Sampler


@datamodel(proto=training_pb.SearchSpec)
class ModelSearch(Configuration):
    """ SearchSpec specifies the configuration for a distributed model search """
    Sampler: SamplerType = SamplerType.TPESearch
    """ The hyper-parameter optimization search method """
    Pruner: PrunerSettings = None
    """
    Pruner specifies the configuration to run a model search using a pruner algorithm. Using a pruning
    algorithm allows you to train a large number of candidate models with a subset of the dataset
    """
    MaxCost: int = 100
    """ The maximum cost that can be incurred before stopping the model search (applicable for deep learning models) """
    MaxTime: int = 30
    """ The maximum number of time, in seconds, that the model search can run for """
    MaxModels: int = 10
    """ The maximum number of candidate models that will be sampled and trained """
    MinScore: float = 0
    """ The minimum score of a candidate model, after which the model search will forcibly stop """
    Trainers: int = 1
    """
    The desired number of trainers that will train candidate models in parallel. The number
    of trainers is restricted based on the allowance provided by the active License
    """
    Test: int = 1
    """
    The number of top candidate models that will be moved to the testing phase once the model search is complete.
    By default, only the best model will be retained
    """
    RetainTop: int = 10
    """
    The number of top candidate models, sorted by their objective score, that will be retained in
    the case that garbage collection is enabled. All other models will be archived
    """
    RetainedFor: int = 60
    """
    The time, in minutes, for which candidate models (excluding the best model) will be
    retained, after which they will be archived
    """
    SearchSpace: AlgorithmSearchSpace = AlgorithmSearchSpace()
    """ SearchSpace specifies the algorithms available to candidate models """
    EarlyStopAfter: int = 0
    """
    The number of new models produced by the search which, if there is no improvement
    in score, the model search will conclude
    """
    KeepOnlyTopModel: bool = True
    Objective: Metric = None
    """ The objective metric that will be measured against all models to evaluate their performance """
    Objective2: Metric = None
    """
    The second objective metric that will be measured and evaluated in tandem with the primary objective.
    The optimizer will attempt to optimize both metrics
    """


@datamodel(proto=training_pb.BaselineSpec)
class BaselineSettings(Configuration):
    """ BaselineSpec specifies the configuration to produce baseline models """
    Enabled: bool = False
    """
    Indicates if baseline models will be produced. Enabling baseline will create a model for each
    algorithm in the parent Study's search space with default hyper-parameters
    """
    Baselines: List[ClassicEstimator] = field(default_factory=lambda : [])
    """ Baselines contains the collection of algorithms to create models for """
    All: bool = False
    """ Indicates if models will be created for all algorithms, ignoring the Baselines field """


@datamodel(proto=training_pb.FeatureEngineeringSearchSpec)
class FeatureEngineeringSearch(Configuration):
    """ FeatureEngineeringSearchSpec specifies the configuration to produce """
    Enabled: bool = True
    """ Indicates if the feature engineering search will be performed """
    ImbalancedHandler: ImbalanceHandling = ImbalanceHandling.ImbalanceAuto
    """ The method to use when handling an imbalanced dataset """
    Estimator: ClassicEstimator = ClassicEstimator.DecisionTreeClassifier
    """ The algorithm to use when evaluating models with different feature engineering pipelines """
    MaxModels: int = 2
    """
    The number of models to sample, after which the feature engineering with the highest
    score will be used with Model resources produced by the primary model search of the parent Study
    """
    MaxTime: int = 3600
    """ The deadline, in seconds, for models produced by the search to be trained """
    MaxTrainers: int = 1
    """
    The desired number of trainers that will train candidate models in parallel. The number
    of trainers is restricted based on the allowance provided by the active License
    """
    SamplePct: int = 100
    """ The number percentage (0 through 100) of the dataset to train models with """
    AutoRemove: bool = True
    """
    Indicates if models produced by the feature engineering search should be
    automatically removed at the conclusion of the search
    """
    Reuse: bool = False
    """
    If true, if a feature engineering pipeline was previously produced for
    the same dataset it will be used as a starting point for the search
    """
    FeatureSelectionTemplate: FeatureSelection = FeatureSelection()
    """
    Specification for feature selection.
    successful study.
    """


@datamodel(proto=training_pb.StudyScheduleSpec)
class StudySchedule(Configuration):
    """ StudyScheduleSpec specifies the parameters for a Study to be executed at a certain time """
    Enabled: bool = False
    """ Indicates if the schedule is enabled """
    StartAt: Time = None
    """ The time at which the Study will begin execution """


@datamodel(proto=training_pb.ModelResult)
class ModelResult(Configuration):
    """ ModelResult contains the records of a single garbage-collected model """
    Name: str = ''
    """ The name of the model """
    Alg: str = ''
    """ The type of estimator of the model """
    Score: float = 0
    """ The objective score of the model """
    Error: bool = False
    """ Indicates if the model experience an error whilst training """
    TrialID: int = 0
    """ The optimizer trial ID of the model """


@datamodel(proto=training_pb.GarbageCollectionSpec)
class GarbageCollection(Configuration):
    """ GarbageCollectionSpec specifies the configuration to garbage-collect unused Model resources """
    CollectAtStudyEnd: bool = True
    """ Indicates if all models of the Study, excluding the highest-performing model, will be archived """
    KeepOnlyBestModelPerAlgorithm: bool = True
    """
    Indicates if multiple models with the same algorithm are produced by the Study, the
    model with the lowest score will be garbage-collected
    """


@datamodel(proto=training_pb.EnsemblesSpec)
class Ensemble(Configuration):
    """ EnsemblesSpec specifies the configuration to produce ensemble models """
    Enabled: bool = False
    """ Indicates if ensemble models will be created """
    VotingEnsemble: bool = False
    """ Indicates if a voting ensemble model will be created """
    StackingEnsemble: bool = True
    """ Indicates if a stacking ensemble model will be created """
    Top: int = 3
    """ The number of top candidate models to include in the ensemble """


@datamodel(proto=training_pb.StudySpec)
class StudySpec(Configuration):
    """ StudySpec defines the desired state of a Study and the parameters for a model search """
    VersionName: str = 'v0.0.1'
    """
    The name of the DataProductVersion which describes the version of the resource
    that exists in the same DataProduct namespace as the resource
    """
    Description: str = ''
    """ The user-provided description of the Study """
    LabRef: ObjectReference = ObjectReference('default-tenant', 'default-lab')
    """
    The reference to the Lab under which the Model resources created by the Study will be trained.
    If unspecified, the default Lab from the parent DataProduct will be used
    """
    DatasetName: str = ''
    """
    The name of the Dataset resource which will be split into training, testing, and
    validation datasets to be used in training
    """
    Task: TaskType = TaskType.AutoDetectTask
    """ The machine learning task type of the Study (i.e. regression, classification) """
    FeSearch: FeatureEngineeringSearch = FeatureEngineeringSearch()
    """ FeatureEngineeringSearch specifies the parameters to perform a feature engineering search """
    Baseline: BaselineSettings = BaselineSettings()
    """ Baseline specifies the parameters to generate baseline (default hyper-parameters) models """
    Search: ModelSearch = ModelSearch()
    """ Search specifies the configuration to perform the model search for the best algorithm and hyper-parameters """
    Ensembles: Ensemble = Ensemble()
    """ Ensembles specifies to parameters to generate ensemble models """
    TrainingTemplate: Training = Training()
    """ TrainingTemplate specifies the configuration to train and evaluate models """
    Schedule: StudySchedule = StudySchedule()
    """ Schedule specifies the configuration to execute the Study at a later date """
    Interpretability: InterpretabilitySpec = InterpretabilitySpec()
    """ Interpretability specifies the parameters to create interpretability visualizations for the final model """
    Aborted: bool = False
    """ Aborted indicates that the execution of the Study and associated Models should be permanently stopped """
    Reported: bool = True
    """ Reported indicates that a report will be generated for the Study """
    Paused: bool = False
    """ Paused indicates that the execution of new workloads associated with the Study should be paused """
    Profiled: bool = True
    """ Profiled indicates that the Study will be profiled after the conclusion of it's model search """
    ModelPublished: bool = False
    """ ModelPublished indicates that a Docker image will be created containing the best model produced by the Study """
    ModelImagePushed: bool = False
    """ ModelImagePushed indicates that if a Docker image of the best model will be pushed to a Docker image registry """
    ModelBenchmarked: bool = True
    ModelExplained: bool = True
    """
    ModelExplained indicates if interpretability diagrams, as specified
    by the Interpretability field, will be produced for the final model
    """
    Fast: bool = False
    """ Fast indicates if Models associated with the Study should skip profiling, explaining, and reporting """
    Location: DataLocation = DataLocation(BucketName='default-minio-bucket')
    """ The data location where Study artifacts (metadata, reports, and model artifacts) generated by the Study will be stored """
    Owner: str = 'no-one'
    """ The name of the Account which created the object, which exists in the same tenant as the object """
    ActiveDeadlineSeconds: int = 600
    """ The deadline for any Jobs associated with the Study to be completed in seconds """
    Compilation: CompilerSettings = None
    """ CompilerSpec specifies the configuration to a compile the best model to a binary (currently unimplemented) """
    Template: bool = False
    """ Indicates if the Study is a template, in which case it will not be executed """
    Flagged: bool = False
    """ Indicates if the Study is flagged """
    Notification: NotificationSettings = NotificationSettings()
    """ The notification specification that determines which notifiers will receive Alerts generated by the object """
    Gc: GarbageCollection = GarbageCollection()
    """ GarbageCollectionSpec specifies the configuration to automatically clean-up unused models """
    Ttl: int = 0
    """ The time-to-live, in seconds, for Model resources produced by the Study """


@datamodel(proto=training_pb.StudyCondition)
class StudyCondition(Configuration):
    """ StudyCondition describes the state of a Study at a certain point """
    Type: StudyConditionType = None
    """ Type of study condition """
    Status: ConditionStatus = None
    """ Status of the condition, one of True, False, Unknown """
    LastTransitionTime: Time = None
    """ Last time the condition transitioned from one status to another """
    Reason: str = ''
    """ The reason for the condition's last transition """
    Message: str = ''
    """ A human-readable message indicating details about the transition """


@datamodel(proto=training_pb.GarbageCollectionStatus)
class GarbageCollectionStatus(Configuration):
    """ GarbageCollectionStatus contains the records for garbage-collected models """
    Collected: int = 0
    """ The number of models that were collected, equal to len(Models) """
    Models: List[ModelResult] = field(default_factory=lambda : [])
    """ The collection of models that were archived """


@datamodel(proto=training_pb.StudyPhaseStatus)
class StudyPhaseStatus(Configuration):
    """ StudyPhaseStatus contains the statistics for a single phase of a Study """
    StartTime: Time = None
    """ The time at which the phase started """
    EndTime: Time = None
    """ The time at which the phase concluded """
    Waiting: int = 0
    """ The number of models pending training """
    Running: int = 0
    """ The number of models currently being trained """
    Failed: int = 0
    """ The number of models that experienced an error whilst training """
    Completed: int = 0
    """ The number of models that have been successfully trained """


@datamodel(proto=training_pb.StudyStatus)
class StudyStatus(ImmutableConfiguration):
    """ StudyStatus defines the observed state of a Study """
    Models: int = 0
    """ Total models created for the study """
    StartTime: Time = None
    """ StartTime represents the time at which the execution of the Study started """
    EndTime: Time = None
    """ EndTime represents the time at which the Study was marked as completed, failed, or aborted """
    BestModel: str = ''
    """ The name of the Model resource which was determined to be the highest-performing """
    BestModelScore: float = 0
    """ The score of the Model resource which was determined to be the highest-performing """
    ProfileUri: str = ''
    """ The URI of the raw profile data produced by the Study """
    ReportName: str = ''
    """ The name of the Report resource produced by the Study """
    Phase: StudyPhase = StudyPhase.ModelPending
    """ The phase of the Study """
    ObservedGeneration: int = 0
    """ ObservedGeneration is the last generation that was acted on """
    TrainDataset: DataLocation = None
    """ TrainDatasetLocation specifies the location of the training dataset """
    TestDataset: DataLocation = None
    """ TestDatasetLocation specifies the location of the testing dataset """
    ValidationDataset: DataLocation = None
    """ ValidationDataset specifies the location of the validation dataset """
    LastModelID: int = 0
    """ The Kubernetes-internal ID of the last Model resource generated by the Study """
    FailureReason: StatusError = None
    """ In the case of failure, the Study resource controller will set this field with a failure reason """
    FailureMessage: str = ''
    """ In the case of failure, the Study resource controller will set this field with a failure message """
    TrainingRows: int = 0
    """ The number of rows in the training dataset """
    TestingRows: int = 0
    """ The number of rows in the testing dataset """
    ValidationRows: int = 0
    """ The number of rows in the validation dataset """
    Progress: int = 0
    """ The progress percentage of the Study, which is derived from the Study's current phase """
    BaselineModel: ClassicEstimator = None
    TrainingDataHash: DataHashes = None
    """
    Sha 256 of the data sig
    TrainingDataHash specifies the hashes for datasets used by the Study
    """
    TriggeredBy: TriggerType = None
    """ The type of trigger which started the Study """
    Logs: OutputLogs = None
    """ Logs specifies the location of logs produced by workloads associated with the Study """
    FeatureEngineering: StudyPhaseStatus = None
    """ FeatureEngineeringStatus contains the status of the feature engineering phase """
    Baseline: StudyPhaseStatus = None
    """ BaselineStatus contains the status of the baseline phase """
    Search: StudyPhaseStatus = None
    """ SearchStatus contains the status of the model search phase """
    Ensemble: StudyPhaseStatus = None
    """ EnsembleStatus contains the status of the ensemble phase """
    Test: StudyPhaseStatus = None
    """ TestStatus contains the status of the testing phase """
    Explain: StudyPhaseStatus = None
    """ ExplainStatus contains the status of the explaining phase """
    LastUpdated: Time = None
    """ The last time the object was updated """
    BestFE: FeatureEngineeringSpec = None
    """ BestFE specifies the best feature engineering pipeline produced by the Study """
    Gc: GarbageCollectionStatus = None
    """ GC specifies the status of garbage collection relevant to the Study """
    Conditions: List[StudyCondition] = field(default_factory=lambda : [])


@datamodel(proto=training_pb.ReportCondition)
class ReportCondition(Configuration):
    """ ReportCondition describes the state of a Report at a certain point. """
    Type: ReportConditionType = None
    """ Type of Report condition """
    Status: ConditionStatus = None
    """ Status of the condition, one of True, False, Unknown """
    LastTransitionTime: Time = None
    """ Last time the condition transitioned from one status to another """
    Reason: str = ''
    """ The reason for the condition's last transition """
    Message: str = ''
    """ A human-readable message indicating details about the transition """


ReportType_ = ReportType


@datamodel(proto=training_pb.ReportSpec)
class ReportSpec(Configuration):
    """ ReportSpec specifies the desired state of a Report """
    VersionName: str = 'v0.0.1'
    """
    The name of the DataProductVersion which describes the version of the resource
    that exists in the same DataProduct namespace as the resource
    """
    EntityRef: ObjectReference = None
    """ EntityRef specifies the entity which the Report references. The supported entities consist of Dataset, Model, and Study resources """
    Location: DataLocation = DataLocation(BucketName='modela')
    """ The location of the flat-file containing the PDF report """
    ReportType: ReportType_ = None
    """ The type of report (e.g. classification model report, study report) """
    Format: ReportFormat = ReportFormat.Pdf
    """ The format of the Report. `pdf` is the only supported type as of the current release """
    NotifierName: str = ''
    """ The name of the Notifier resource which Alerts created by the Report will be forwarded to """
    Owner: str = 'no-one'
    """ The name of the Account which created the object, which exists in the same tenant as the object """
    Resources: Workload = Workload('memory-large')
    """ Resources specifies the resource requirements that will be allocated to the report generation workload """
    ActiveDeadlineSeconds: int = 600
    """ The deadline for any Jobs associated with the Report to be completed in seconds """


@datamodel(proto=training_pb.ReportStatus)
class ReportStatus(Configuration):
    """ ReportStatus defines the observed state of a Report """
    StartTime: Time = None
    """ StartTime represents the time at which the execution of the Report started """
    EndTime: Time = None
    """ EndTime represents the time at which the Report was marked as completed, failed, or aborted """
    Phase: ReportPhase = ReportPhase.Pending
    """ The phase of the Report """
    Uri: str = ''
    """ The URI to the flat-file report within the VirtualBucket specified by the Report """
    ObservedGeneration: int = 0
    """ ObservedGeneration is the last generation that was acted on """
    FailureReason: StatusError = None
    """ In the case of failure, the Report resource controller will set this field with a failure reason """
    FailureMessage: str = ''
    """ In the case of failure, the Report resource controller will set this field with a failure message """
    Logs: OutputLogs = None
    """ Logs specifies the location of logs produced by workloads associated with the Report """
    LastUpdated: Time = None
    """ The last time the object was updated """
    Conditions: List[ReportCondition] = field(default_factory=lambda : [])


@datamodel(proto=training_pb.ModelAutobuilderCondition)
class ModelAutobuilderCondition(Configuration):
    """ ModelAutobuilderCondition describes the state of a ModelAutobuilder at a certain point """
    Type: ModelAutobuilderConditionType = None
    """ Type of ModelAutobuilder condition """
    Status: ConditionStatus = None
    """ Status of the condition, one of True, False, Unknown """
    LastTransitionTime: Time = None
    """ Last time the condition transitioned from one status to another """
    Reason: str = ''
    """ The reason for the condition's last transition """
    Message: str = ''
    """ A human-readable message indicating details about the transition """


DataSourceTemplate = DataSourceSpec
DatasetType_ = DatasetType


@datamodel(proto=training_pb.ModelAutobuilderSpec)
class ModelAutobuilderSpec(Configuration):
    """ ModelAutobuilderSpec define the desired state of a ModelAutobuilder """
    DataProductName: str = ''
    """ The name of the DataProduct namespace that the resource exists under """
    DataProductVersionName: str = ''
    """
    The name of the DataProductVersion which describes the version of the resource
    that exists in the same DataProduct namespace as the resource
    """
    DatasourceName: str = ''
    """
    DataSourceName is the name of an existing DataSource resource which will be used as the schema for the ModelAutoBuilder's Dataset.
    If empty, a DataSource will be automatically created based on the data specified by the Location field
    """
    DatasetName: str = ''
    """
    The name of an existing Dataset resource, or the name of the Dataset resource that will be created
    based on the data specified by the Location field, which will be used to train models
    """
    Location: DataLocation = None
    """ The location for data that will be saved in a Dataset resource to train models with """
    Task: TaskType = None
    """ The machine learning task type relevant to the dataset (i.e. regression, classification) """
    Objective: Metric = None
    """ The objective metric that will be measured against trained models to evaluate their performance """
    TargetColumn: str = ''
    """ The name of the column within the dataset that contains the label(s) to be predicted """
    MaxTime: int = 600
    """ The deadline for models to complete training, in seconds """
    MaxModels: int = 10
    """ The number of candidate models that will be sampled and trained """
    AccessMethod: AccessType = AccessType.ClusterIP
    """
    The Kubernetes-native access method which specifies how the Predictor created by the ModelAutobuilder will be exposed.
    See https://modela.ai/docs/docs/serving/production/#access-method for a detailed description of each access type
    """
    AutoScale: bool = False
    """ Indicates if the Predictor created by the ModelAutobuilder will automatically scale to traffic """
    Dataapp: bool = False
    """ Indicates if the ModelAutobuilder will create a DataApp resource to serve the highest-performing model that was trained """
    DataSourceSpec: DataSourceSpec = None
    """
    DataSourceSpec specifies the full specification of the DataSource resource that will be created by the ModelAutobuilder.
    If empty, the ModelAutobuilder will attempt to infer the schema of the data specified by the Location field
    """
    Trainers: int = 1
    """
    The desired number of trainers that will train candidate models in parallel. The number
    of trainers is restricted based on the allowance provided by the active License
    """
    Sampler: SamplerType = SamplerType.TPESearch
    """ The hyper-parameter optimization search method """
    Aborted: bool = False
    """ Aborted indicates that the execution of the ModelAutobuilder and any associated workloads should be permanently stopped """
    Owner: str = 'no-one'
    """ The name of the Account which created the object, which exists in the same tenant as the object """
    Resources: Workload = None
    """ Resources specifies the resource requirements that will be allocated to dataset and model training Jobs """
    FeatureEngineering: bool = False
    """ Indicates if feature engineering will be performed prior to the primary model search """
    FeatureSelection: bool = False
    """ Indicates if feature selection will be performed prior to the primary model search """
    ServingSiteRef: ObjectReference = ObjectReference(Namespace='default-tenant', Name='default-serving-site')
    """
    The reference to the ServingSite where the Predictor created by the ModelAutobuilder will be hosted.
    If unspecified, the default ServingSite from the parent DataProduct will be used
    """
    LabRef: ObjectReference = ObjectReference(Namespace='default-tenant', Name='default-lab')
    """
    The reference to the Lab under which Dataset and Study resources created by the ModelAutobuilder will be trained.
    If unspecified, the default Lab from the parent DataProduct will be used
    """
    DatasetType: DatasetType_ = DatasetType_.Tabular
    """ The type of dataset which was uploaded. `tabular` is the only supported type as of the current release """


@datamodel(proto=training_pb.ModelAutobuilderStatus)
class ModelAutobuilderStatus(Configuration):
    """ ModelAutobuilderStatus define the observed state of a ModelAutobuilder """
    FlatFileName: str = ''
    """ The name of the flat-file generated for the associated Dataset """
    DataSourceName: str = ''
    """ The name of the DataSource associated with resource """
    DatasetName: str = ''
    """ The name of the Dataset associated with the resource """
    StudyName: str = ''
    """ The name of the Study associated with the resource """
    BestModelName: str = ''
    """ The name of the highest-performing Model resource produced as a result of the associated Study resource """
    PredictorName: str = ''
    """ The name of the Predictor associated with the resource """
    ImageRepoName: str = ''
    Phase: ModelAutobuilderPhase = ModelAutobuilderPhase.Pending
    """ The phase of the ModelAutobuilder """
    Rows: int = 0
    """ The number of rows observed in the Dataset associated with the resource """
    Cols: int = 0
    """ The number of columns observed in the Dataset associated with the resource """
    FileSize: int = 0
    """ The size of the raw data in the Dataset associated with the resource """
    Models: int = 0
    """ The number of total Model resources created by the associated Study resource """
    TrainedModels: int = 0
    """ The number of successfully trained Model resources created by the associated Study resource """
    BestModelScore: float = 0
    """ The highest score out of all Models created by the associated Study resource """
    Estimator: ClassicalEstimatorSpec = None
    """ The estimator specification for the highest-performing Model resource """
    StartTime: Time = None
    """ StartTime represents the time at which the execution of the ModelAutobuilder started """
    EndTime: Time = None
    """ EndTime represents the time at which the ModelAutobuilder was marked as completed, failed, or aborted """
    ObservedGeneration: int = 0
    """ ObservedGeneration is the last generation that was acted on """
    FailureReason: StatusError = None
    """ In the case of failure, the ModelAutobuilder resource controller will set this field with a failure reason """
    FailureMessage: str = ''
    """ In the case of failure, the ModelAutobuilder resource controller will set this field with a failure message """
    LastUpdated: Time = None
    """ The last time the object was updated """
    Conditions: List[ModelAutobuilderCondition] = field(default_factory=lambda : [])



import grpc
from github.com.metaprov.modelaapi.pkg.apis.data.v1alpha1.generated_pb2 import Feature as MDFeature
from github.com.metaprov.modelaapi.services.feature.v1.feature_pb2_grpc import FeatureServiceStub
from github.com.metaprov.modelaapi.services.feature.v1.feature_pb2 import CreateFeatureRequest, \
    UpdateFeatureRequest, \
    DeleteFeatureRequest, GetFeatureRequest, ListFeaturesRequest

from modela.Resource import Resource
from modela.ModelaException import ModelaException
from typing import List, Union


class Feature(Resource):
    def __init__(self, item: MDFeature = MDFeature(), client=None, namespace="", name="", version=Resource.DefaultVersion):
        super().__init__(item, client, namespace=namespace, name=name, version=version)


class FeatureClient:
    def __init__(self, stub, modela):
        self.modela = modela
        self.__stub: FeatureServiceStub = stub

    def create(self, feature: Feature) -> bool:
        request = CreateFeatureRequest()
        request.feature.CopyFrom(feature.raw_message)
        try:
            response = self.__stub.CreateFeature(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def update(self, feature: Feature) -> bool:
        request = UpdateFeatureRequest()
        request.feature.CopyFrom(feature.raw_message)
        try:
            self.__stub.UpdateFeature(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def get(self, namespace: str, name: str) -> Union[Feature, bool]:
        request = GetFeatureRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.GetFeature(request)
            return Feature(response.feature, self)
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def delete(self, namespace: str, name: str) -> bool:
        request = DeleteFeatureRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.DeleteFeature(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def list(self, namespace: str) -> Union[List[Feature], bool]:
        request = ListFeaturesRequest()
        request.namespace = namespace
        try:
            response = self.__stub.ListFeatures(request)
            return [Feature(item, self) for item in response.features.items]
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False





import grpc
from github.com.metaprov.modelaapi.pkg.apis.training.v1alpha1.generated_pb2 import Model as MDModel
from github.com.metaprov.modelaapi.services.model.v1.model_pb2_grpc import ModelServiceStub
from github.com.metaprov.modelaapi.services.model.v1.model_pb2 import \
    DeleteModelRequest, GetModelRequest, ListModelsRequest, \
    AbortModelRequest, CompileModelRequest, DeployModelRequest, DownloadModelRequest, \
    PauseModelRequest, GetModelProfileRequest, PublishModelRequest, ResumeModelRequest, TestModelRequest
from tabulate import tabulate

from modela.data.Dataset import Dataset
from modela.Resource import Resource
from modela.ModelaException import ModelaException
from typing import List, Union

from modela.training.Report import Report
from modela.training.common import ModelPhase, TaskType, Metric
from modela.training.models import ModelSpec, ModelStatus, ModelProfile


class Model(Resource):
    def __init__(self, item: MDModel = MDModel(), client=None, namespace="", name=""):
        """
        The Model resource is a machine learning model generated by the Modela data plane. Models are immutable and
        cannot be changed once created.

        :param client: The Model client repository, which can be obtained through an instance of Modela.
        :param namespace: The target namespace of the resource.
        :param name: The name of the resource.
        """
        super().__init__(item, client, namespace=namespace, name=name)

    @property
    def spec(self) -> ModelSpec:
        return ModelSpec().copy_from(self._object.spec)

    @property
    def status(self) -> ModelStatus:
        return ModelStatus().copy_from(self._object.status)

    def default(self):
        raise TypeError("Model {0} was not found; model resources cannot be created.".format(self.name))

    def submit(self):
        raise TypeError("Model resources cannot be created.")

    def update(self):
        raise TypeError("Model resources cannot be updated.")

    def profile(self) -> ModelProfile:
        if hasattr(self, "_client"):
            profile = self._client.profile(self.namespace, self.name)
        else:
            raise AttributeError("Object has no client repository")

        return ModelProfile().copy_from(profile)

    def abort(self):
        if hasattr(self, "_client"):
            self._client.abort(self.namespace, self.name)
        else:
            raise AttributeError("Object has no client repository")

    def pause(self):
        if hasattr(self, "_client"):
            self._client.pause(self.namespace, self.name)
        else:
            raise AttributeError("Object has no client repository")

    def resume(self):
        if hasattr(self, "_client"):
            self._client.resume(self.namespace, self.name)
        else:
            raise AttributeError("Object has no client repository")

    def compile(self, target, compiler):
        if hasattr(self, "_client"):
            self._client.compile(self.namespace, self.name, target, compiler)
        else:
            raise AttributeError("Object has no client repository")

    def publish(self):
        if hasattr(self, "_client"):
            self._client.publish(self.namespace, self.name)
        else:
            raise AttributeError("Object has no client repository")

    def test(self):
        if hasattr(self, "_client"):
            self._client.test(self.namespace, self.name)
        else:
            raise AttributeError("Object has no client repository")

    def download(self) -> bytes:
        if hasattr(self, "_client"):
            return self._client.download(self.namespace, self.name)
        else:
            raise AttributeError("Object has no client repository")

    def deploy(self, predictor: str, replicas: int = 0, traffic: int = 0, role: str = ""):
        if hasattr(self, "_client"):
            self._client.deploy(self.namespace, self.name, predictor, replicas, traffic, role)
        else:
            raise AttributeError("Object has no client repository")

    @property
    def test_prediction(self) -> str:
        """ Generate a default prediction payload for the model. """
        return self.dataset.test_prediction

    @property
    def report(self) -> Report:
        """ Return the report associated with the model, if it exists. """
        if hasattr(self, "_client"):
            if self._object.status.reportName != "":
                return self._client.modela.Report(namespace=self.namespace, name=self._object.status.reportName)
            else:
                print("Model {0} has no report. Call Model.test() to create one.")
        else:
            raise AttributeError("Object has no client repository")


    @property
    def dataset(self) -> Dataset:
        """ Return the dataset associated with the model """
        if hasattr(self, "_client"):
            return self._client.modela.Dataset(namespace=self.namespace, name=self.spec.DatasetName)
        else:
            raise AttributeError("Object has no client repository")


    @property
    def phase(self) -> ModelPhase:
        return self.status.Phase

    @property
    def details(self) -> str:
        """
        Generate a table about the details of the model.

        :return: A table of the model's metrics and hyperparameters
        """

        return "Estimator: {0} | Trial #: {1}\n".format(self.spec.Estimator.AlgorithmName, self.spec.TrialID) + \
                self.metrics + "\n" + self.hyperparameters

    @property
    def hyperparameters(self) -> str:
        """
        Generate a table with the hyperparameters of the model.
        """
        hyper_table = []
        for parameters in self.spec.Estimator.Parameters:
            hyper_table.append([parameters.Name, parameters.Value])

        return tabulate(hyper_table, tablefmt='psql', headers=['Hyperparameter', 'Value'])

    @property
    def metrics(self) -> str:
        """
        Generate a table with the metrics of the model.
        """
        test_table, headers, headers_as_metric, status = [], ['Metric'], [], self.status
        if self.spec.Task == TaskType.BinaryClassification:
            metrics = ['Accuracy', 'Auc', 'F1', 'Log Loss', 'Precision', 'Recall']
            headers_as_metric = [Metric.Accuracy, Metric.RocAuc, Metric.F1Binary, Metric.LogLoss, Metric.PrecisionBinary, Metric.RecallBinary]
            test_table = [[metric] for metric in metrics]
        elif self.spec.Task == TaskType.MultiClassification:
            metrics = ['Accuracy', 'F1 Micro', 'F1 Macro', 'F1 Weighted', 'Precision Macro', 'Precision Micro', 'Precision Weighted',
                       'Recall Macro', 'Recall Micro', 'Recall Weighted']
            headers_as_metric = [Metric.Accuracy, Metric.F1Micro, Metric.F1Macro, Metric.F1Weighted, Metric.PrecisionMacro, Metric.PrecisionMicro,
                                 Metric.PrecisionWeighted, Metric.RecallMacro, Metric.RecallMicro, Metric.RecallWeighted]
            test_table = [[metric] for metric in metrics]
        elif self.spec.Task == TaskType.Regression:
            headers = []

        if len(self.status.Cv) > 0:
            headers.append("Validation")
            for idx, metric in enumerate(headers_as_metric):
                test_table[idx].append([x for x in status.Cv if x.Metric == metric][0].Value)

        if len(self.status.Train) > 0:
            headers.append("Train")
            for idx, metric in enumerate(headers_as_metric):
                test_table[idx].append([x for x in status.Train if x.Metric == metric][0].Value)

        if len(self.status.Test) > 0:
            headers.append("Test")
            for idx, metric in enumerate(headers_as_metric):
                test_table[idx].append([x for x in self.status.Test if x.Metric == metric][0].Value)

        return tabulate(test_table, tablefmt='psql', headers=headers)

    def get_cv_metric(self, metric: Metric) -> float:
        out_metric = [x for x in self.status.Cv if x.Metric == metric]
        if len(out_metric) == 0:
            raise TypeError("{0} model does not have the {1} metric.".format(self.spec.Task.name, metric.name))

        return out_metric[0].Value

    def get_train_metric(self, metric: Metric) -> float:
        out_metric = [x for x in self.status.Train if x.Metric == metric]
        if len(out_metric) == 0:
            raise TypeError("{0} model does not have the {1} metric..".format(self.spec.Task.name, metric.name))

        return out_metric[0].Value

    def get_test_metric(self, metric: Metric) -> float:
        out_metric = [x for x in self.status.Test if x.Metric == metric]
        if len(out_metric) == 0:
            raise TypeError("{0} model does not have the {1} metric.".format(self.spec.Task.name, metric.name))

        return out_metric[0].Value

    def __repr__(self):
        return "<{0} model at {1}/{2}>".format(self.spec.Estimator.AlgorithmName, self.namespace, self.name)



class ModelClient:
    def __init__(self, stub, modela):
        self.__stub: ModelServiceStub = stub
        self.modela = modela

    def create(self, model: Model) -> bool:
        raise TypeError("Modela does not support the creation of custom models in this release.")
        # request = CreateModelRequest()
        # request.model.CopyFrom(model.raw_message)
        # try:
        #    response = self.__stub.CreateModel(request)
        #    return True
        # except grpc.RpcError as err:
        #    error = err

        # ModelaException.process_error(error)

    def update(self, model: Model) -> bool:
        raise TypeError("Model resources cannot be changed after creation.")
        # request = UpdateModelRequest()
        # request.model.CopyFrom(model.raw_message)
        # try:
        #    self.__stub.UpdateModel(request)
        #    return True
        # except grpc.RpcError as err:
        #    error = err

        # ModelaException.process_error(error)
        # return False

    def get(self, namespace: str, name: str) -> Union[Model, bool]:
        request = GetModelRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.GetModel(request)
            return Model(response.model, self)
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def delete(self, namespace: str, name: str) -> bool:
        request = DeleteModelRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.DeleteModel(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def list(self, namespace: str, labels: dict = None) -> Union[List[Model], bool]:
        request = ListModelsRequest()
        request.namespace = namespace
        if labels is not None:
            request.labels.update(labels)

        try:
            response = self.__stub.ListModels(request)
            return [Model(item, self) for item in response.models.items]
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def abort(self, namespace: str, name: str) -> bool:
        request = AbortModelRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.AbortModel(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def compile(self, namespace: str, name: str, target: str, compiler: str) -> bool:
        request = CompileModelRequest()
        request.namespace = namespace
        request.name = name
        request.target = target
        request.compiler = compiler
        try:
            response = self.__stub.CompileModel(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def deploy(self, namespace: str, name: str, predictor: str, replicas: int = 0,
               traffic: int = 0, role: str = "") -> bool:
        request = DeployModelRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.DeployModel(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def download(self, namespace: str, name: str) -> bytes:
        request = DownloadModelRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.DownloadModel(request)
            return response.raw
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def pause(self, namespace: str, name: str) -> bool:
        request = PauseModelRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.PauseModel(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def resume(self, namespace: str, name: str) -> bool:
        request = ResumeModelRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.ResumeModel(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def test(self, namespace: str, name: str) -> bool:
        request = TestModelRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.TestModel(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def profile(self, namespace: str, name: str) -> bool:
        request = GetModelProfileRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.GetModelProfile(request)
            return response.profile
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def publish(self, namespace: str, name: str) -> bool:
        request = PublishModelRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.PublishModel(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False





import subprocess
import socket
from contextlib import closing

from typing import Callable

from github.com.metaprov.modelaapi.services.account.v1.account_pb2 import AccountLoginRequest
from github.com.metaprov.modelaapi.services.datapipelinerun.v1 import datapipelinerun_pb2_grpc
from github.com.metaprov.modelaapi.services.datapipeline.v1 import datapipeline_pb2_grpc
from github.com.metaprov.modelaapi.services.dataproduct.v1 import dataproduct_pb2_grpc
from github.com.metaprov.modelaapi.services.dataproductversion.v1 import dataproductversion_pb2_grpc
from github.com.metaprov.modelaapi.services.dataset.v1 import dataset_pb2_grpc
from github.com.metaprov.modelaapi.services.datasource.v1 import datasource_pb2_grpc
from github.com.metaprov.modelaapi.services.entity.v1 import entity_pb2_grpc
from github.com.metaprov.modelaapi.services.featurehistogram.v1 import featurehistogram_pb2_grpc
from github.com.metaprov.modelaapi.services.featurepipelinerun.v1 import featurepipelinerun_pb2_grpc
from github.com.metaprov.modelaapi.services.featurepipeline.v1 import featurepipeline_pb2_grpc
from github.com.metaprov.modelaapi.services.feature.v1 import feature_pb2_grpc
from github.com.metaprov.modelaapi.services.featureset.v1 import featureset_pb2_grpc
from github.com.metaprov.modelaapi.services.labelingpipelinerun.v1 import labelingpipelinerun_pb2_grpc
from github.com.metaprov.modelaapi.services.labelingpipeline.v1 import labelingpipeline_pb2_grpc
from github.com.metaprov.modelaapi.services.reciperun.v1 import reciperun_pb2_grpc
from github.com.metaprov.modelaapi.services.recipe.v1 import recipe_pb2_grpc
from github.com.metaprov.modelaapi.services.sqlquery.v1 import sqlquery_pb2_grpc
from github.com.metaprov.modelaapi.services.sqlqueryrun.v1 import sqlqueryrun_pb2_grpc
from github.com.metaprov.modelaapi.services.webrequestrun.v1 import webrequestrun_pb2_grpc
from github.com.metaprov.modelaapi.services.webrequest.v1 import webrequest_pb2_grpc
from github.com.metaprov.modelaapi.services.cronprediction.v1 import cronprediction_pb2_grpc
from github.com.metaprov.modelaapi.services.fileservices.v1 import fileservices_pb2_grpc
from github.com.metaprov.modelaapi.services.curtain.v1 import curtain_pb2_grpc
from github.com.metaprov.modelaapi.services.dataapp.v1 import dataapp_pb2_grpc
from github.com.metaprov.modelaapi.services.prediction.v1 import prediction_pb2_grpc
from github.com.metaprov.modelaapi.services.predictor.v1 import predictor_pb2_grpc
from github.com.metaprov.modelaapi.services.account.v1 import account_pb2_grpc
from github.com.metaprov.modelaapi.services.alert.v1 import alert_pb2_grpc
from github.com.metaprov.modelaapi.services.apitoken.v1 import apitoken_pb2_grpc
from github.com.metaprov.modelaapi.services.attachment.v1 import attachment_pb2_grpc
from github.com.metaprov.modelaapi.services.commit.v1 import commit_pb2_grpc
from github.com.metaprov.modelaapi.services.connection.v1 import connection_pb2_grpc
from github.com.metaprov.modelaapi.services.lab.v1 import lab_pb2_grpc
from github.com.metaprov.modelaapi.services.license.v1 import license_pb2_grpc
from github.com.metaprov.modelaapi.services.modelasystem.v1 import modelasystem_pb2_grpc
from github.com.metaprov.modelaapi.services.notifier.v1 import notifier_pb2_grpc
from github.com.metaprov.modelaapi.services.servingsite.v1 import servingsite_pb2_grpc
from github.com.metaprov.modelaapi.services.tenant.v1 import tenant_pb2_grpc
from github.com.metaprov.modelaapi.services.virtualbucket.v1 import virtualbucket_pb2_grpc
from github.com.metaprov.modelaapi.services.virtualcluster.v1 import virtualcluster_pb2_grpc
from github.com.metaprov.modelaapi.services.virtualvolume.v1 import virtualvolume_pb2_grpc
from github.com.metaprov.modelaapi.services.userroleclass.v1 import userroleclass_pb2_grpc
from github.com.metaprov.modelaapi.services.meeting.v1 import meeting_pb2_grpc
from github.com.metaprov.modelaapi.services.postmortem.v1 import postmortem_pb2_grpc
from github.com.metaprov.modelaapi.services.review.v1 import review_pb2_grpc
from github.com.metaprov.modelaapi.services.runbook.v1 import runbook_pb2_grpc
from github.com.metaprov.modelaapi.services.todo.v1 import todo_pb2_grpc
from github.com.metaprov.modelaapi.services.cronreport.v1 import cronreport_pb2_grpc
from github.com.metaprov.modelaapi.services.modelautobuilder.v1 import modelautobuilder_pb2_grpc
from github.com.metaprov.modelaapi.services.modelcompilerrun.v1 import modelcompilerrun_pb2_grpc
from github.com.metaprov.modelaapi.services.modelpipelinerun.v1 import modelpipelinerun_pb2_grpc
from github.com.metaprov.modelaapi.services.modelpipeline.v1 import modelpipeline_pb2_grpc
from github.com.metaprov.modelaapi.services.model.v1 import model_pb2_grpc
from github.com.metaprov.modelaapi.services.notebookrun.v1 import notebookrun_pb2_grpc
from github.com.metaprov.modelaapi.services.notebook.v1 import notebook_pb2_grpc
from github.com.metaprov.modelaapi.services.report.v1 import report_pb2_grpc
from github.com.metaprov.modelaapi.services.study.v1 import study_pb2_grpc

from modela.data.DataPipelineRun import *
from modela.data.DataPipeline import *
from modela.data.DataProduct import *
from modela.data.DataProductVersion import *
from modela.data.Dataset import *
from modela.data.DataSource import *
from modela.data.Entity import *
from modela.data.FeatureHistogram import *
from modela.data.FeaturePipelineRun import *
from modela.data.FeaturePipeline import *
from modela.data.Feature import *
from modela.data.Featureset import *
from modela.data.LabelingPipelineRun import *
from modela.data.LabelingPipeline import *
from modela.data.RecipeRun import *
from modela.data.Recipe import *
from modela.data.SqlQuery import *
from modela.data.SqlQueryRun import *
from modela.data.WebRequestRun import *
from modela.data.WebRequest import *
from modela.inference.CronPrediction import *
from modela.inference.Curtain import *
from modela.inference.DataApp import *
from modela.inference.Prediction import *
from modela.inference.Predictor import *
from modela.inference.InferenceService import *
from modela.infra.FileService import *
from modela.infra.Account import *
from modela.infra.Alert import *
from modela.infra.ApiToken import *
from modela.infra.Attachment import *
from modela.infra.Commit import *
from modela.infra.Connection import *
from modela.infra.Lab import *
from modela.infra.License import *
from modela.infra.ModelaSystem import *
from modela.infra.Notifier import *
from modela.infra.ServingSite import *
from modela.infra.Tenant import *
from modela.infra.VirtualBucket import *
from modela.infra.VirtualCluster import *
from modela.infra.VirtualVolume import *
from modela.infra.UserRoleClass import *
from modela.team.Meeting import *
from modela.team.PostMortem import *
from modela.team.Review import *
from modela.team.RunBook import *
from modela.team.Todo import *
from modela.training.CronReport import *
from modela.training.ModelAutobuilder import *
from modela.training.ModelCompilerRun import *
from modela.training.ModelPipelineRun import *
from modela.training.ModelPipeline import *
from modela.training.Model import *
from modela.training.NotebookRun import *
from modela.training.Notebook import *
from modela.training.Report import *
from modela.training.Study import *
from grpc_interceptor import ClientCallDetails, ClientInterceptor


class AuthClientInterceptor(ClientInterceptor):
    def __init__(self, token: str):
        self.token = token

    def intercept(
            self,
            method: Callable,
            request_or_iterator,
            call_details: grpc.ClientCallDetails,
    ):
        new_details = ClientCallDetails(
            call_details.method,
            call_details.timeout,
            [("authorization", self.token)],
            call_details.credentials,
            call_details.wait_for_ready,
            call_details.compression,
        )

        return method(request_or_iterator, new_details)


class Modela:
    """
    The Modela class provides an interface for connecting to your Modela API Gateway and accessing the Modela custom
    resources on your cluster. The class provides you with a directory for all resource API clients, and functions to
    fetch or create each resource.
    """

    def __init__(
            self,
            host="localhost",
            port=3000,
            username: str = "admin",
            password: str = "admin",
            secure=False,
            tls_cert=None,
            tenant="default-tenant",
            port_forward=False,
    ):
        """
        Connect to the Modela API gateway.

        :param host: The DNS name or IP that hosts the API gateway.
        :param port: The port which exposes the API gateway.
        :param username: Username for your Modela account that has permissions to use the specified tenant.
        :param password: Password for your Modela account.
        :param secure: If connecting through gRPC ingress, secure must be enabled and a TLS public key must be supplied.
        :param tls_cert: The TLS public key of the ingress that exposes the API gateway gRPC service.
        :param tenant: The tenant that Modela SDK objects will utilize by default (default: `default-tenant`)
        :param port_forward: If enabled, the SDK will attempt to port forward the API gateway using kubectl. Kubectl
            must be installed and must be connected to a cluster with Modela installed.
        """

        self.tenant = tenant
        if port_forward:
            with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
                s.bind(('', 0))
                s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                port = s.getsockname()[1]

            self.pf_process = subprocess.Popen(
                "kubectl port-forward svc/modela-api-gateway %d:8080 -n modela-system" % port,
                shell=True, stderr=subprocess.STDOUT)

            time.sleep(1 / 4)
            secure, host = False, "localhost"

        if secure:
            with open(tls_cert, 'rb') as f:
                credentials = grpc.ssl_channel_credentials(f.read())

            self._channel = grpc.secure_channel(f'{host}', credentials)
        else:
            if port != None:
                self._channel = grpc.insecure_channel(f'{host}:{port}')
            else:
                self._channel = grpc.insecure_channel(f'{host}')

        # if interceptors:
        #     self._channel = grpc.intercept_channel(  # type: ignore
        #         self._channel, *interceptors
        #     )

        self.__account_stub = account_pb2_grpc.AccountServiceStub(self._channel)
        self.__account_client = AccountClient(self.__account_stub, self)

        if username is not None:
            login_request = AccountLoginRequest(namespace=tenant, name=username, password=password)
            token = self.__account_stub.Login(login_request).token
            self._channel = grpc.intercept_channel(self._channel, AuthClientInterceptor(token))

        self.__account_stub = account_pb2_grpc.AccountServiceStub(self._channel)
        self.__account_client = AccountClient(self.__account_stub, self)

        self.__fileservice_stub = fileservices_pb2_grpc.FileServicesServiceStub(self._channel)
        self.__fileservice_client = FileService(self.__fileservice_stub)

        self.__datapipelinerun_stub = datapipelinerun_pb2_grpc.DataPipelineRunServiceStub(self._channel)
        self.__datapipelinerun_client = DataPipelineRunClient(self.__datapipelinerun_stub, self)

        self.__datapipeline_stub = datapipeline_pb2_grpc.DataPipelineServiceStub(self._channel)
        self.__datapipeline_client = DataPipelineClient(self.__datapipeline_stub, self)

        self.__dataproduct_stub = dataproduct_pb2_grpc.DataProductServiceStub(self._channel)
        self.__dataproduct_client = DataProductClient(self.__dataproduct_stub, self)

        self.__dataproductversion_stub = dataproductversion_pb2_grpc.DataProductVersionServiceStub(self._channel)
        self.__dataproductversion_client = DataProductVersionClient(self.__dataproductversion_stub, self)

        self.__dataset_stub = dataset_pb2_grpc.DatasetServiceStub(self._channel)
        self.__dataset_client = DatasetClient(self.__dataset_stub, self)

        self.__datasource_stub = datasource_pb2_grpc.DataSourceServiceStub(self._channel)
        self.__datasource_client = DataSourceClient(self.__datasource_stub, self)

        self.__entity_stub = entity_pb2_grpc.EntityServiceStub(self._channel)
        self.__entity_client = EntityClient(self.__entity_stub, self)

        self.__featurehistogram_stub = featurehistogram_pb2_grpc.FeatureHistogramServiceStub(self._channel)
        self.__featurehistogram_client = FeatureHistogramClient(self.__featurehistogram_stub, self)

        self.__featurepipelinerun_stub = featurepipelinerun_pb2_grpc.FeaturePipelineRunServiceStub(self._channel)
        self.__featurepipelinerun_client = FeaturePipelineRunClient(self.__featurepipelinerun_stub, self)

        self.__featurepipeline_stub = featurepipeline_pb2_grpc.FeaturePipelineServiceStub(self._channel)
        self.__featurepipeline_client = FeaturePipelineClient(self.__featurepipeline_stub, self)

        self.__feature_stub = feature_pb2_grpc.FeatureServiceStub(self._channel)
        self.__feature_client = FeatureClient(self.__feature_stub, self)

        self.__featureset_stub = featureset_pb2_grpc.FeaturesetServiceStub(self._channel)
        self.__featureset_client = FeaturesetClient(self.__featureset_stub, self)

        self.__labelingpipelinerun_stub = labelingpipelinerun_pb2_grpc.LabelingPipelineRunServiceStub(self._channel)
        self.__labelingpipelinerun_client = LabelingPipelineRunClient(self.__labelingpipelinerun_stub, self)

        self.__labelingpipeline_stub = labelingpipeline_pb2_grpc.LabelingPipelineServiceStub(self._channel)
        self.__labelingpipeline_client = LabelingPipelineClient(self.__labelingpipeline_stub, self)

        self.__reciperun_stub = reciperun_pb2_grpc.RecipeRunServiceStub(self._channel)
        self.__reciperun_client = RecipeRunClient(self.__reciperun_stub, self)

        self.__recipe_stub = recipe_pb2_grpc.RecipeServiceStub(self._channel)
        self.__recipe_client = RecipeClient(self.__recipe_stub, self)

        self.__sqlquery_stub = sqlquery_pb2_grpc.SqlQueryServiceStub(self._channel)
        self.__sqlquery_client = SqlQueryClient(self.__sqlquery_stub, self)

        self.__sqlqueryrun_stub = sqlqueryrun_pb2_grpc.SqlQueryRunServiceStub(self._channel)
        self.__sqlqueryrun_client = SqlQueryRunClient(self.__sqlqueryrun_stub, self)

        self.__webrequestrun_stub = webrequestrun_pb2_grpc.WebRequestRunServiceStub(self._channel)
        self.__webrequestrun_client = WebRequestRunClient(self.__webrequestrun_stub, self)

        self.__webrequest_stub = webrequest_pb2_grpc.WebRequestServiceStub(self._channel)
        self.__webrequest_client = WebRequestClient(self.__webrequest_stub, self)

        self.__cronprediction_stub = cronprediction_pb2_grpc.CronPredictionServiceStub(self._channel)
        self.__cronprediction_client = CronPredictionClient(self.__cronprediction_stub, self)

        self.__curtain_stub = curtain_pb2_grpc.CurtainServiceStub(self._channel)
        self.__curtain_client = CurtainClient(self.__curtain_stub, self)

        self.__dataapp_stub = dataapp_pb2_grpc.DataAppServiceStub(self._channel)
        self.__dataapp_client = DataAppClient(self.__dataapp_stub, self)

        self.__prediction_stub = prediction_pb2_grpc.PredictionServiceStub(self._channel)
        self.__prediction_client = PredictionClient(self.__prediction_stub, self)

        self.__predictor_stub = predictor_pb2_grpc.PredictorServiceStub(self._channel)
        self.__predictor_client = PredictorClient(self.__predictor_stub, self)

        self.__alert_stub = alert_pb2_grpc.AlertServiceStub(self._channel)
        self.__alert_client = AlertClient(self.__alert_stub, self)

        self.__apitoken_stub = apitoken_pb2_grpc.ApiTokenServiceStub(self._channel)
        self.__apitoken_client = ApiTokenClient(self.__apitoken_stub, self)

        self.__attachment_stub = attachment_pb2_grpc.AttachmentServiceStub(self._channel)
        self.__attachment_client = AttachmentClient(self.__attachment_stub, self)

        self.__commit_stub = commit_pb2_grpc.CommitServiceStub(self._channel)
        self.__commit_client = CommitClient(self.__commit_stub, self)

        self.__connection_stub = connection_pb2_grpc.ConnectionServiceStub(self._channel)
        self.__connection_client = ConnectionClient(self.__connection_stub, self)

        self.__lab_stub = lab_pb2_grpc.LabServiceStub(self._channel)
        self.__lab_client = LabClient(self.__lab_stub, self)

        self.__license_stub = license_pb2_grpc.LicenseServiceStub(self._channel)
        self.__license_client = LicenseClient(self.__license_stub, self)

        self.__modelasystem_stub = modelasystem_pb2_grpc.ModelaSystemServiceStub(self._channel)
        self.__modelasystem_client = ModelaSystemClient(self.__modelasystem_stub, self)

        self.__notifier_stub = notifier_pb2_grpc.NotifierServiceStub(self._channel)
        self.__notifier_client = NotifierClient(self.__notifier_stub, self)

        self.__servingsite_stub = servingsite_pb2_grpc.ServingSiteServiceStub(self._channel)
        self.__servingsite_client = ServingSiteClient(self.__servingsite_stub, self)

        self.__tenant_stub = tenant_pb2_grpc.TenantServiceStub(self._channel)
        self.__tenant_client = TenantClient(self.__tenant_stub, self)

        self.__virtualbucket_stub = virtualbucket_pb2_grpc.VirtualBucketServiceStub(self._channel)
        self.__virtualbucket_client = VirtualBucketClient(self.__virtualbucket_stub, self)

        self.__virtualcluster_stub = virtualcluster_pb2_grpc.VirtualClusterServiceStub(self._channel)
        self.__virtualcluster_client = VirtualClusterClient(self.__virtualcluster_stub, self)

        self.__virtualvolume_stub = virtualvolume_pb2_grpc.VirtualVolumeServiceStub(self._channel)
        self.__virtualvolume_client = VirtualVolumeClient(self.__virtualvolume_stub, self)

        self.__userroleclass_stub = userroleclass_pb2_grpc.UserRoleClassServiceStub(self._channel)
        self.__userroleclass_client = UserRoleClassClient(self.__userroleclass_stub, self)

        self.__meeting_stub = meeting_pb2_grpc.MeetingServiceStub(self._channel)
        self.__meeting_client = MeetingClient(self.__meeting_stub, self)

        self.__postmortem_stub = postmortem_pb2_grpc.PostMortemServiceStub(self._channel)
        self.__postmortem_client = PostMortemClient(self.__postmortem_stub, self)

        self.__review_stub = review_pb2_grpc.ReviewServiceStub(self._channel)
        self.__review_client = ReviewClient(self.__review_stub, self)

        self.__runbook_stub = runbook_pb2_grpc.RunBookServiceStub(self._channel)
        self.__runbook_client = RunBookClient(self.__runbook_stub, self)

        self.__todo_stub = todo_pb2_grpc.TodoServiceStub(self._channel)
        self.__todo_client = TodoClient(self.__todo_stub, self)

        self.__cronreport_stub = cronreport_pb2_grpc.CronReportServiceStub(self._channel)
        self.__cronreport_client = CronReportClient(self.__cronreport_stub, self)

        self.__modelautobuilder_stub = modelautobuilder_pb2_grpc.ModelAutobuilderServiceStub(self._channel)
        self.__modelautobuilder_client = ModelAutobuilderClient(self.__modelautobuilder_stub, self)

        self.__modelcompilerrun_stub = modelcompilerrun_pb2_grpc.ModelCompilerRunServiceStub(self._channel)
        self.__modelcompilerrun_client = ModelCompilerRunClient(self.__modelcompilerrun_stub, self)

        self.__modelpipelinerun_stub = modelpipelinerun_pb2_grpc.ModelPipelineRunServiceStub(self._channel)
        self.__modelpipelinerun_client = ModelPipelineRunClient(self.__modelpipelinerun_stub, self)

        self.__modelpipeline_stub = modelpipeline_pb2_grpc.ModelPipelineServiceStub(self._channel)
        self.__modelpipeline_client = ModelPipelineClient(self.__modelpipeline_stub, self)

        self.__model_stub = model_pb2_grpc.ModelServiceStub(self._channel)
        self.__model_client = ModelClient(self.__model_stub, self)

        self.__notebookrun_stub = notebookrun_pb2_grpc.NotebookRunServiceStub(self._channel)
        self.__notebookrun_client = NotebookRunClient(self.__notebookrun_stub, self)

        self.__notebook_stub = notebook_pb2_grpc.NotebookServiceStub(self._channel)
        self.__notebook_client = NotebookClient(self.__notebook_stub, self)

        self.__report_stub = report_pb2_grpc.ReportServiceStub(self._channel)
        self.__report_client = ReportClient(self.__report_stub, self)

        self.__study_stub = study_pb2_grpc.StudyServiceStub(self._channel)
        self.__study_client = StudyClient(self.__study_stub, self)

    @property
    def Accounts(self) -> AccountClient:
        return self.__account_client

    def Account(self, namespace="", name="") -> Account:
        return Account(MDAccount(), self.Accounts, namespace=namespace, name=name)

    @property
    def DataProducts(self) -> DataProductClient:
        return self.__dataproduct_client

    def DataProduct(self, namespace="", name="", serving_site: Union[ServingSite, str] = None,
                    lab: Union[Lab, str] = None,
                    public: bool = None, task_type: TaskType = None, default_training_workload: Workload = None,
                    default_serving_workload: Workload = None, default_bucket: str = None,
                    notification_settings: NotificationSettings = None,
                    permissions: Permissions = None) -> DataProduct:
        """
        :param namespace: The target namespace of the resource.
        :param name: The name of the resource.
        :param serving_site: The default Serving Site of the Data Product.
        :param lab: The default Lab of the Data Product.
        :param public: If enabled, the Data Product will be publicly accessible by all users without permissions.
        :param task_type: The default task type for child resources of the Data Product.
        :param default_training_workload: The default workload for training Jobs under the Data Product
        :param default_serving_workload: The default workload for model serving Jobs under the Data Product.
        :param default_bucket: The default bucket used for child resources of the Data Product.
        :param notification_settings: The default notification settings used for child resources of the Data Product.
        :param permissions: The permission specification that dictates which users can access the resources under
          the Data Product and what actions they can perform.
        """
        return DataProduct(MDDataProduct(), self.DataProducts, namespace, name, serving_site, lab, public,
                           task_type, default_training_workload, default_serving_workload, default_bucket,
                           notification_settings, permissions)

    @property
    def DataSources(self) -> DataSourceClient:
        return self.__datasource_client

    def DataSource(self, namespace="", name="", version=Resource.DefaultVersion, bucket: str = "default-minio-bucket",
                   infer_file: str = None, infer_dataframe: pandas.DataFrame = None, infer_bytes: bytes = None,
                   target_column: str = "", file_type: FlatFileType = None,
                   task_type: TaskType = None, csv_config: CsvFileFormat = None,
                   excel_config: ExcelNotebookFormat = None) -> DataSource:
        """
        :param namespace: The target namespace of the resource.
        :param name: The name of the resource.
        :param version: The version of the resource.
        :param bucket: If data is provided for inference then a bucket must be provided.
        :param infer_file: If specified, the SDK will attempt read a file with the given path and will upload it to
            analyse the columns and generate a schema that will be applied to the resource.
        :param infer_dataframe: If specified, the Pandas DataFrame will be serialized and uploaded to analyse
            the columns and generate a schema that will be applied to the resource.
        :param infer_bytes: If specified, the raw byte data will be uploaded to analyse
            the columns and generate a schema that will be applied to the resource.
        :param target_column: The name of the target column used when training a model. This parameter only has effect
            when data is uploaded to infer a schema.
        :param file_type: The file type of raw data, used when ingesting a Dataset from a file, or creating a data snapshot
            from a database source. If inferring from a dataframe, the file type will default to CSV.
        :param task_type: The target task type in relation to the data being used.
        :param csv_config: The CSV file format of the raw data.
        :param excel_config: The Excel file format of the raw data.
        """

        return DataSource(MDDataSource(), self.DataSources, namespace, name, version, bucket, infer_file,
                          infer_dataframe, infer_bytes, target_column, file_type, task_type, csv_config, excel_config)

    @property
    def DataPipelineRuns(self) -> DataPipelineRunClient:
        return self.__datapipelinerun_client

    def DataPipelineRun(self, namespace="", name="") -> DataPipelineRun:
        return DataPipelineRun(MDDataPipelineRun(), self.DataPipelineRuns, namespace, name)

    @property
    def DataPipelines(self) -> DataPipelineClient:
        return self.__datapipeline_client

    def DataPipeline(self, namespace="", name="") -> DataPipeline:
        return DataPipeline(MDDataPipeline(), self.DataPipelines, namespace, name)

    @property
    def DataProductVersions(self) -> DataProductVersionClient:
        return self.__dataproductversion_client

    def DataProductVersion(self, namespace="", name="", baseline: bool = False,
                           previous_version: str = None) -> DataProductVersion:
        """
        Fetch or create a new Data Product Version resource

        :param namespace: The target namespace of the resource.
        :param name: The name of the resource.
        :param baseline: If this version is baseline, then the objects of previous version will be garbage collected.
        :param previous_version: The name of the previous version.
        """
        return DataProductVersion(MDDataProductVersion(), self.DataProductVersions, namespace, name, baseline,
                                  previous_version)

    @property
    def Datasets(self) -> DatasetClient:
        return self.__dataset_client

    def Dataset(self, namespace="", name="", version=Resource.DefaultVersion, gen_datasource: bool = False,
                lab: Union[ObjectReference, Lab, str] = "default-lab",
                target_column: str = None, datasource: Union[DataSource, str] = "",
                bucket: Union[VirtualBucket, str] = "default-minio-bucket",
                dataframe: pandas.DataFrame = None, data_file: str = None, data_bytes: bytes = None,
                workload: Workload = Workload("general-large"), fast: bool = False,
                sample: SampleSettings = None, task_type: TaskType = None,
                notification: NotificationSettings = None) -> Dataset:
        """
        Fetch or create a new Dataset resource

        :param client: The Dataset client repository, which can be obtained through an instance of Modela.
        :param namespace: The target namespace of the resource.
        :param name: The name of the resource.
        :param lab: The object reference, Lab object, or lab name under the tenant of the resource for which all
            Dataset-related workloads will be performed under.
        :param gen_datasource: If true, a Datasource resource will be created from the uploaded dataset and applied to
            the Dataset resource.
        :param target_column: If gen_datasource is enabled, then the target column of the data source must be specified.
        :param datasource: If specified as a string, the SDK will attempt to find a Data Source resource with the given name.
            If specified as a Data Source, or if one was found with the given name, it will be applied to the Dataset.
        :param bucket: The bucket which the raw dataset data will be uploaded to.
        :param dataframe: The Pandas Dataframe will be serialized and uploaded for ingestion with the Dataset resource.
        :param data_file: The file path which will be read and uploaded for ingestion with the Dataset resource.
        :param data_bytes: The raw data as bytes that will be uploaded for  ingestion with the Dataset resource.
        :param workload: The resource requirements which will be allocated for Dataset ingestion.
        :param fast: If enabled, the Dataset will skip validation, profiling, and reporting.
        :param sample: The sample settings of the dataset, which if enabled will select only a portion of the Dataset
          for further processing.
        :param task_type: The target task type in relation to the data being used.
        :param notification: The notification settings, which if enabled will forward events about this resource to a notifier.
        """
        return Dataset(MDDataset(), self.Datasets, namespace, name, version, lab, gen_datasource, target_column,
                       datasource, bucket, dataframe, data_file, data_bytes, workload, fast,
                       sample, task_type, notification)

    @property
    def FileService(self) -> FileService:
        return self.__fileservice_client

    @property
    def Entities(self) -> EntityClient:
        return self.__entity_client

    def Entity(self, namespace="", name="") -> Entity:
        return Entity(MDEntity(), self.Entities, namespace, name)

    @property
    def FeatureHistograms(self) -> FeatureHistogramClient:
        return self.__featurehistogram_client

    def FeatureHistogram(self, namespace="", name="") -> FeatureHistogram:
        return FeatureHistogram(MDFeatureHistogram(), self.FeatureHistograms, namespace, name)

    @property
    def FeaturePipelineRuns(self) -> FeaturePipelineClient:
        return self.__featurepipelinerun_client

    def FeaturePipelineRun(self, namespace="", name="") -> FeaturePipelineRun:
        return FeaturePipelineRun(MDFeaturePipelineRun(), self.FeaturePipelineRuns, namespace, name)

    @property
    def FeaturePipelines(self) -> FeaturePipelineClient:
        return self.__featurepipeline_client

    def FeaturePipeline(self, namespace="", name="") -> FeaturePipeline:
        return FeaturePipeline(MDFeaturePipeline(), self.FeaturePipelines, namespace, name)

    @property
    def Features(self) -> FeatureClient:
        return self.__feature_client

    def Feature(self, namespace="", name="") -> Feature:
        return Feature(MDFeature(), self.Features, namespace, name)

    @property
    def Featuresets(self) -> FeaturesetClient:
        return self.__featureset_client

    def Featureset(self, namespace="", name="") -> Featureset:
        return Featureset(MDFeatureset(), self.Featuresets, namespace, name)

    @property
    def LabelingPipelineRuns(self) -> LabelingPipelineRunClient:
        return self.__labelingpipelinerun_client

    def LabelingPipelineRun(self, namespace="", name="") -> LabelingPipelineRun:
        return LabelingPipelineRun(MDLabelingPipelineRun(), self.LabelingPipelineRuns, namespace, name)

    @property
    def LabelingPipelines(self) -> LabelingPipelineClient:
        return self.__labelingpipeline_client

    def LabelingPipeline(self, namespace="", name="") -> LabelingPipeline:
        return LabelingPipeline(MDLabelingPipeline(), self.LabelingPipelines, namespace, name)

    @property
    def RecipeRuns(self) -> RecipeRunClient:
        return self.__reciperun_client

    def RecipeRun(self, namespace="", name="") -> RecipeRun:
        return RecipeRun(MDRecipeRun(), self.RecipeRuns, namespace, name)

    @property
    def Recipes(self) -> RecipeClient:
        return self.__recipe_client

    def Recipe(self, namespace="", name="") -> Recipe:
        return Recipe(MDRecipe(), self.Recipes, namespace, name)

    @property
    def SqlQuerys(self) -> SqlQueryClient:
        return self.__sqlquery_client

    def SqlQuery(self, namespace="", name="") -> SqlQuery:
        return SqlQuery(MDSqlQuery(), self.SqlQuerys, namespace, name)

    @property
    def SqlQueryRuns(self) -> SqlQueryRunClient:
        return self.__sqlqueryrun_client

    def SqlQueryRun(self, namespace="", name="") -> SqlQueryRun:
        return SqlQueryRun(MDSqlQueryRun(), self.SqlQueryRuns, namespace, name)

    @property
    def WebRequestRuns(self) -> WebRequestRunClient:
        return self.__webrequestrun_client

    def WebRequestRun(self, namespace="", name="") -> WebRequestRun:
        return WebRequestRun(MDWebRequestRun(), self.WebRequestRuns, namespace, name)

    @property
    def WebRequests(self) -> WebRequestClient:
        return self.__webrequest_client

    def WebRequest(self, namespace="", name="") -> WebRequest:
        return WebRequest(MDWebRequest(), self.WebRequests, namespace, name)

    @property
    def CronPredictions(self) -> CronPredictionClient:
        return self.__cronprediction_client

    def CronPrediction(self, namespace="", name="") -> CronPrediction:
        return CronPrediction(MDCronPrediction(), self.CronPredictions, namespace, name)

    @property
    def Curtains(self) -> CurtainClient:
        return self.__curtain_client

    def Curtain(self, namespace="", name="") -> Curtain:
        return Curtain(MDCurtain(), self.Curtains, namespace, name)

    @property
    def DataApps(self) -> DataAppClient:
        return self.__dataapp_client

    def DataApp(self, namespace="", name="") -> DataApp:
        return DataApp(MDDataApp(), self.DataApps, namespace, name)

    @property
    def Predictions(self) -> PredictionClient:
        return self.__prediction_client

    def Prediction(self, namespace="", name="") -> Prediction:
        return Prediction(MDPrediction(), self.Predictions, namespace, name)

    @property
    def Predictors(self) -> PredictorClient:
        return self.__predictor_client

    def Predictor(self, namespace="", name="", version=Resource.DefaultVersion,
                  serving_site: Union[ObjectReference, ServingSite, str] = "default-serving-site",
                  model: Union[Model, str] = None, models: List[ModelDeploymentSpec] = None, port: int = 3000,
                  path: str = None, access_type: AccessType = None, replicas: int = 0, autoscale: bool = False,
                  workload: Workload = None) -> Predictor:
        """
        :param namespace: The target namespace of the resource.
        :param name: The name of the resource.
        :param version: The version of the resource.
        :param serving_site: The object reference, Serving Site object, or name under the tenant of the resource for which the
            predictor will be deployed under.
        :param model: The Model object or name that the predictor will serve predictions for. This parameter
            will instantiate the predictor with a singe default deployment specification for the model.
        :param models: The list of model deployment specifications that will be applied to the predictor.
        :param port: The port of the predictor, which if applicable will expose the service internally or externally.
        :param path: The path that the predictor will expose a service on. If this parameter is not specified, then
            a path will be generated based on the access type and predictor name.
        :param access_type: The access type of the Predictor. See https://www.modela.ai/docs/docs/serving/production/
            for documentation on how different access types expose the Predictor service
        :param replicas: The amount of replicas for the predictor, which if greater than zero will serve the prediction
            service on multiple pods.
        :param autoscale: If set to true, the predictor's deployment will scale based on the amount of incoming traffic
            to the service.
        :param workload: The workload specification which determines the resources which will be allocated to the
            prediction service.
        """
        return Predictor(MDPredictor(), self.Predictors, namespace, name, version, serving_site, model, models,
                         port, path, access_type, replicas, autoscale, workload)

    @property
    def Alerts(self) -> AlertClient:
        return self.__alert_client

    def Alert(self, namespace="", name="") -> Alert:
        return Alert(MDAlert(), self.Alerts, namespace, name)

    @property
    def ApiTokens(self) -> ApiTokenClient:
        return self.__apitoken_client

    def ApiToken(self, namespace="", name="") -> ApiToken:
        return ApiToken(MDApiToken(), self.ApiTokens, namespace, name)

    @property
    def Attachments(self) -> AttachmentClient:
        return self.__attachment_client

    def Attachment(self, namespace="", name="") -> Attachment:
        return Attachment(MDAttachment(), self.Attachments, namespace, name)

    @property
    def Commits(self) -> CommitClient:
        return self.__commit_client

    def Commit(self, namespace="", name="") -> Commit:
        return Commit(MDCommit(), self.Commits, namespace, name)

    @property
    def Connections(self) -> ConnectionClient:
        return self.__connection_client

    def Connection(self, namespace="", name="") -> Connection:
        return Connection(MDConnection(), self.Connections, namespace, name)

    @property
    def Labs(self) -> LabClient:
        return self.__lab_client

    def Lab(self, namespace="", name="") -> Lab:
        return Lab(MDLab(), self.Labs, namespace, name)

    @property
    def Licenses(self) -> LicenseClient:
        return self.__license_client

    def License(self, namespace="", name="") -> License:
        return License(MDLicense(), self.Licenses, namespace, name)

    @property
    def ModelaSystems(self) -> ModelaSystemClient:
        return self.__modelasystem_client

    def ModelaSystem(self, namespace="", name="") -> ModelaSystem:
        return ModelaSystem(MDModelaSystem(), self.ModelaSystems, namespace, name)

    @property
    def Notifiers(self) -> NotifierClient:
        return self.__notifier_client

    def Notifier(self, namespace="", name="") -> Notifier:
        return Notifier(MDNotifier(), self.Notifiers, namespace, name)

    @property
    def ServingSites(self) -> ServingSiteClient:
        return self.__servingsite_client

    def ServingSite(self, namespace="", name="") -> ServingSite:
        return ServingSite(MDServingSite(), self.ServingSites, namespace, name)

    @property
    def Tenants(self) -> TenantClient:
        return self.__tenant_client

    def Tenant(self, namespace="", name="") -> Tenant:
        return Tenant(MDTenant(), self.Tenants, namespace, name)

    @property
    def VirtualBuckets(self) -> VirtualBucketClient:
        return self.__virtualbucket_client

    def VirtualBucket(self, namespace="", name="") -> VirtualBucket:
        return VirtualBucket(MDVirtualBucket(), self.VirtualBuckets, namespace, name)

    @property
    def VirtualClusters(self) -> VirtualClusterClient:
        return self.__virtualcluster_client

    def VirtualCluster(self, namespace="", name="") -> VirtualCluster:
        return VirtualCluster(MDVirtualCluster(), self.VirtualClusters, namespace, name)

    @property
    def VirtualVolumes(self) -> VirtualVolumeClient:
        return self.__virtualvolume_client

    def VirtualVolume(self, namespace="", name="") -> VirtualVolume:
        return VirtualVolume(MDVirtualVolume(), self.VirtualVolumes, namespace, name)

    @property
    def UserRoleClasses(self) -> UserRoleClassClient:
        return self.__userroleclass_client

    def UserRoleClass(self, namespace="", name="", rules: List[Rule] = None) -> UserRoleClass:
        """
        :param namespace: The target namespace of the resource.
        :param name: The name of the resource.
        :param rules: The list of rules which any Account associated with the User Role Class resource may perform.
        """
        return UserRoleClass(MDUserRoleClass(), self.UserRoleClasses, namespace, name, rules)

    @property
    def Meetings(self) -> MeetingClient:
        return self.__meeting_client

    def Meeting(self, namespace="", name="") -> Meeting:
        return Meeting(MDMeeting(), self.Meetings, namespace, name)

    @property
    def PostMortems(self) -> PostMortemClient:
        return self.__postmortem_client

    def PostMortem(self, namespace="", name="") -> PostMortem:
        return PostMortem(MDPostMortem(), self.PostMortems, namespace, name)

    @property
    def Reviews(self) -> ReviewClient:
        return self.__review_client

    def Review(self, namespace="", name="") -> Review:
        return Review(MDReview(), self.Reviews, namespace, name)

    @property
    def RunBooks(self) -> RunBookClient:
        return self.__runbook_client

    def RunBook(self, namespace="", name="") -> RunBook:
        return RunBook(MDRunBook(), self.RunBooks, namespace, name)

    @property
    def Todos(self) -> TodoClient:
        return self.__todo_client

    def Todo(self, namespace="", name="") -> Todo:
        return Todo(MDTodo(), self.Todos, namespace, name)

    @property
    def CronReports(self) -> CronReportClient:
        return self.__cronreport_client

    def CronReport(self, namespace="", name="") -> CronReport:
        return CronReport(MDCronReport(), self.CronReports, namespace, name)

    @property
    def ModelAutobuilders(self) -> ModelAutobuilderClient:
        return self.__modelautobuilder_client

    def ModelAutobuilder(self, namespace="", name="", version=Resource.DefaultVersion,
                         lab: Union[ObjectReference, Lab, str] = "default-lab",
                         serving_site: Union[ObjectReference, ServingSite, str] = "default-serving-site",
                         task_type: TaskType = TaskType.BinaryClassification, workload: Workload = Workload("general-large"),
                         bucket: Union[ObjectReference, VirtualBucket, str] = "default-minio-bucket",
                         dataframe: pandas.DataFrame = None, data_file: str = None, data_bytes: bytes = None,
                         target_column: str = None, objective: Metric = Metric.Accuracy,
                         feature_selection: bool = False, feature_engineering: bool = False, max_models: int = 1,
                         max_time: int = 512, trainers: int = 1, predictor_access_type: AccessType = AccessType.ClusterIP,
                         predictor_autoscale: bool = False, create_data_app: bool = False) -> ModelAutobuilder:
        """
        :param namespace: The target namespace of the resource.
        :param name: The name of the resource.
        :param version: The version of the resource.
        :param lab: The object reference, Lab object, or lab name under the tenant of the resource for which all
            data science workloads will be performed under.
        :param serving_site: The object reference, Serving Site object, or name under the tenant of the resource for which
            the created Predictor will be deployed under.
        :param task_type: The target task type in relation to the data being used.
        :param workload: The workload specification which determines the resources which will be allocated for training
            and serving workloads.
        :param bucket: The bucket which the raw dataset data will be uploaded to.
        :param dataframe: The Pandas Dataframe will be serialized and uploaded to create a new Dataset resource.
        :param data_file: The file path which will be read and uploaded to create a new Dataset resource.
        :param data_bytes: The raw data as bytes that will be uploaded to create a new Dataset resource.
        :param target_column: The target column of the Dataset to be created.
        :param objective: The objective metric to be used when determining the best model from the Study.
        :param feature_selection: If True feature selection will be performed on the Dataset before the model search.
        :param feature_engineering: If True feature engineering will be performed on the Dataset before the model search.
        :param max_models: The maximum number of models to sample before selecting the best model.
        :param max_time: The maximum amount of time in seconds which the model search can run for.
        :param trainers: The number of parallel trainers to allocate during the model search
        :param predictor_access_type: The access type of the Predictor to be created. See https://www.modela.ai/docs/docs/serving/production/
            for documentation on how different access types expose the Predictor service
        :param predictor_autoscale: If true the created Predictor will automatically scale for traffic
        :param create_data_app: If true a Data Application will be created to serve a live model dashboard
        """
        return ModelAutobuilder(MDModelAutobuilder(), self.ModelAutobuilders, namespace, name, version, lab,
                                serving_site, task_type, workload, bucket, dataframe, data_file, data_bytes,
                                target_column, objective, feature_selection, feature_engineering, max_models,
                                max_time, trainers, predictor_access_type, predictor_autoscale, create_data_app)

    @property
    def ModelCompilerRuns(self) -> ModelCompilerRunClient:
        return self.__modelcompilerrun_client

    def ModelCompilerRun(self, namespace="", name="") -> ModelCompilerRun:
        return ModelCompilerRun(MDModelCompilerRun(), self.ModelCompilerRuns, namespace, name)

    @property
    def ModelPipelineRuns(self) -> ModelPipelineRunClient:
        return self.__modelpipelinerun_client

    def ModelPipelineRun(self, namespace="", name="") -> ModelPipelineRun:
        return ModelPipelineRun(MDModelPipelineRun(), self.ModelPipelineRuns, namespace, name)

    @property
    def ModelPipelines(self) -> ModelPipelineClient:
        return self.__modelpipeline_client

    def ModelPipeline(self, namespace="", name="") -> ModelPipeline:
        return ModelPipeline(MDModelPipeline(), self.ModelPipelines, namespace, name)

    @property
    def Models(self) -> ModelClient:
        return self.__model_client

    def Model(self, namespace="", name="") -> Model:
        return Model(MDModel(), self.Models, namespace, name)

    @property
    def NotebookRuns(self) -> NotebookRunClient:
        return self.__notebookrun_client

    def NotebookRun(self, namespace="", name="") -> NotebookRun:
        return NotebookRun(MDNotebookRun(), self.NotebookRuns, namespace, name)

    @property
    def Notebooks(self) -> NotebookClient:
        return self.__notebook_client

    def Notebook(self, namespace="", name="") -> Notebook:
        return Notebook(MDNotebook(), self.Notebooks, namespace, name)

    @property
    def Reports(self) -> ReportClient:
        return self.__report_client

    def Report(self, namespace="", name="") -> Report:
        """
        :param namespace: The target namespace of the resource.
        :param name: The name of the resource.
        :param entity: An object reference to an existing resource on the cluster.
        :param report_type: The type of report being generated; this should correspond with the type of entity being
            referenced (e.g. RegressionModelReport for a regression Model resource)
        """
        return Report(MDReport(), self.Reports, namespace, name)

    @property
    def Studies(self) -> StudyClient:
        return self.__study_client

    def Study(self, namespace="", name="", version=Resource.DefaultVersion, dataset: Union[str, Dataset] = "",
              lab: Union[ObjectReference, Lab, str] = "default-lab", bucket: Union[VirtualBucket, str] = None,
              objective: Metric = Metric.Accuracy, search: ModelSearch = None,
              fe_search: FeatureEngineeringSearch = None, baseline: BaselineSettings = None,
              ensemble: Ensemble = None, trainer_template: Training = None, interpretability: Interpretability = None,
              schedule: StudySchedule = None, notification: NotificationSettings = None, garbage_collect: bool = True,
              keep_best_models: bool = True, timeout: int = 600, template: bool = False) -> Study:
        """
        :param namespace: The target namespace of the resource.
        :param name: The name of the resource.
        :param version: The version of the resource.
        :param dataset: If specified as a string, the SDK will attempt to find a Dataset resource with the given name.
            If specified as a Dataset object, or if one was found with the given name, it will be used in the Study.
        :param lab: The object reference, Lab object, or lab name under the default-tenant for which all Study-related
            workloads will be performed under.
        :param bucket: The Bucket object or name of the bucket which will store the Study artifacts
        :param objective: The objective metric relevant to the task type.
        :param search: The search parameters define how many models to sample
        :param fe_search: The feature engineering search parameters of the Study
        :param baseline: The baseline settings for the Study, which if enabled will train an unoptimized model of each
            algorithm type for benchmarking.
        :param ensemble: The ensemble settings for the Study, which if enabled will combine the top estimators of
            the study after the initial model search.
        :param trainer_template: The training template for each model created by the Study.
        :param interpretability: The interpretability configuration for the Study, which specifies what type of
            model explainability plots will be generated from Shap values if they are computed
        :param schedule: The schedule for the study to run chronically
        :param notification: The notification settings, which if enabled will forward events about this resource to a notifier.
        :param garbage_collect: If enabled, models which did not move past the testing stage will be garbage collected by
            the system.
        :param keep_best_models: If enabled, the best models from each algorithm will not be garbage collected.
        :param timeout: The timeout in seconds after which the Study will fail.
        :param template: If the Study is a template it will not start a search and can only be used as a template for
            other studies.
        """
        return Study(MDStudy(), self.Studies, namespace, name, version, dataset, lab, bucket, objective,
                     search, fe_search, baseline, ensemble, trainer_template, interpretability, schedule,
                     notification, garbage_collect, keep_best_models, timeout, template)

    def close(self):
        if hasattr(self, 'pf_process'):
            self.pf_process.kill()

        if self._channel:
            self._channel.close()

    def __del__(self):
        self.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self.close()



import grpc
from github.com.metaprov.modelaapi.pkg.apis.infra.v1alpha1.generated_pb2 import Lab as MDLab
from github.com.metaprov.modelaapi.services.lab.v1.lab_pb2_grpc import LabServiceStub
from github.com.metaprov.modelaapi.services.lab.v1.lab_pb2 import CreateLabRequest, \
    UpdateLabRequest, \
    DeleteLabRequest, GetLabRequest, ListLabsRequest

from modela.Resource import Resource
from modela.ModelaException import ModelaException
from typing import List, Union


class Lab(Resource):
    def __init__(self, item: MDLab = MDLab(), client=None, namespace="", name=""):
        super().__init__(item, client, namespace=namespace, name=name)


class LabClient:
    def __init__(self, stub, modela):
        self.modela = modela
        self.__stub: LabServiceStub = stub

    def create(self, lab: Lab) -> bool:
        request = CreateLabRequest()
        request.lab.CopyFrom(lab.raw_message)
        try:
            response = self.__stub.CreateLab(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def update(self, lab: Lab) -> bool:
        request = UpdateLabRequest()
        request.lab.CopyFrom(lab.raw_message)
        try:
            self.__stub.UpdateLab(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def get(self, namespace: str, name: str) -> Union[Lab, bool]:
        request = GetLabRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.GetLab(request)
            return Lab(response.lab, self)
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def delete(self, namespace: str, name: str) -> bool:
        request = DeleteLabRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.DeleteLab(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def list(self, namespace: str) -> Union[List[Lab], bool]:
        request = ListLabsRequest()
        request.namespace = namespace
        try:
            response = self.__stub.ListLabs(request)
            return [Lab(item, self) for item in response.labs.items]
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False





import grpc
from github.com.metaprov.modelaapi.pkg.apis.data.v1alpha1.generated_pb2 import SqlQuery as MDSqlQuery
from github.com.metaprov.modelaapi.services.sqlquery.v1.sqlquery_pb2_grpc import SqlQueryServiceStub
from github.com.metaprov.modelaapi.services.sqlquery.v1.sqlquery_pb2 import CreateSqlQueryRequest, \
    UpdateSqlQueryRequest, \
    DeleteSqlQueryRequest, GetSqlQueryRequest, ListSqlQuerysRequest

from modela.Resource import Resource
from modela.ModelaException import ModelaException
from typing import List, Union


class SqlQuery(Resource):
    def __init__(self, item: MDSqlQuery = MDSqlQuery(), client=None, namespace="", name="", version=Resource.DefaultVersion):
        super().__init__(item, client, namespace=namespace, name=name, version=version)


class SqlQueryClient:
    def __init__(self, stub, modela):
        self.modela = modela
        self.__stub: SqlQueryServiceStub = stub

    def create(self, sqlquery: SqlQuery) -> bool:
        request = CreateSqlQueryRequest()
        request.sqlquery.CopyFrom(sqlquery.raw_message)
        try:
            response = self.__stub.CreateSqlQuery(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def update(self, sqlquery: SqlQuery) -> bool:
        request = UpdateSqlQueryRequest()
        request.sqlquery.CopyFrom(sqlquery.raw_message)
        try:
            self.__stub.UpdateSqlQuery(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def get(self, namespace: str, name: str) -> Union[SqlQuery, bool]:
        request = GetSqlQueryRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.GetSqlQuery(request)
            return SqlQuery(response.sqlquery, self)
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def delete(self, namespace: str, name: str) -> bool:
        request = DeleteSqlQueryRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.DeleteSqlQuery(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def list(self, namespace: str) -> Union[List[SqlQuery], bool]:
        request = ListSqlQuerysRequest()
        request.namespace = namespace
        try:
            response = self.__stub.ListSqlQuerys(request)
            return [SqlQuery(item, self) for item in response.sqlquerys.items]
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False





import grpc
from github.com.metaprov.modelaapi.pkg.apis.inference.v1alpha1.generated_pb2 import Curtain as MDCurtain
from github.com.metaprov.modelaapi.services.curtain.v1.curtain_pb2_grpc import CurtainServiceStub
from github.com.metaprov.modelaapi.services.curtain.v1.curtain_pb2 import CreateCurtainRequest, \
    UpdateCurtainRequest, \
    DeleteCurtainRequest, GetCurtainRequest, ListCurtainsRequest

from modela.Resource import Resource
from modela.ModelaException import ModelaException
from typing import List, Union


class Curtain(Resource):
    def __init__(self, item: MDCurtain = MDCurtain(), client=None, namespace="", name="", version=Resource.DefaultVersion):
        super().__init__(item, client, namespace=namespace, name=name, version=version)


class CurtainClient:
    def __init__(self, stub, modela):
        self.modela = modela
        self.__stub: CurtainServiceStub = stub

    def create(self, curtain: Curtain) -> bool:
        request = CreateCurtainRequest()
        request.curtain.CopyFrom(curtain.raw_message)
        try:
            response = self.__stub.CreateCurtain(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def update(self, curtain: Curtain) -> bool:
        request = UpdateCurtainRequest()
        request.curtain.CopyFrom(curtain.raw_message)
        try:
            self.__stub.UpdateCurtain(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def get(self, namespace: str, name: str) -> Union[Curtain, bool]:
        request = GetCurtainRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.GetCurtain(request)
            return Curtain(response.curtain, self)
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def delete(self, namespace: str, name: str) -> bool:
        request = DeleteCurtainRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.DeleteCurtain(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def list(self, namespace: str) -> Union[List[Curtain], bool]:
        request = ListCurtainsRequest()
        request.namespace = namespace
        try:
            response = self.__stub.ListCurtains(request)
            return [Curtain(item, self) for item in response.curtains.items]
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False





import grpc
from github.com.metaprov.modelaapi.pkg.apis.infra.v1alpha1.generated_pb2 import ServingSite as MDServingSite
from github.com.metaprov.modelaapi.services.servingsite.v1.servingsite_pb2_grpc import ServingSiteServiceStub
from github.com.metaprov.modelaapi.services.servingsite.v1.servingsite_pb2 import CreateServingSiteRequest, \
    UpdateServingSiteRequest, \
    DeleteServingSiteRequest, GetServingSiteRequest, ListServingSitesRequest

from modela.Resource import Resource
from modela.ModelaException import ModelaException
from typing import List, Union


class ServingSite(Resource):
    def __init__(self, item: MDServingSite = MDServingSite(), client=None, namespace="", name=""):
        super().__init__(item, client, namespace=namespace, name=name)


class ServingSiteClient:
    def __init__(self, stub, modela):
        self.modela = modela
        self.__stub: ServingSiteServiceStub = stub

    def create(self, servingsite: ServingSite) -> bool:
        request = CreateServingSiteRequest()
        request.servingsite.CopyFrom(servingsite.raw_message)
        try:
            response = self.__stub.CreateServingSite(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def update(self, servingsite: ServingSite) -> bool:
        request = UpdateServingSiteRequest()
        request.servingsite.CopyFrom(servingsite.raw_message)
        try:
            self.__stub.UpdateServingSite(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def get(self, namespace: str, name: str) -> Union[ServingSite, bool]:
        request = GetServingSiteRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.GetServingSite(request)
            return ServingSite(response.servingsite, self)
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def delete(self, namespace: str, name: str) -> bool:
        request = DeleteServingSiteRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.DeleteServingSite(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def list(self, namespace: str) -> Union[List[ServingSite], bool]:
        request = ListServingSitesRequest()
        request.namespace = namespace
        try:
            response = self.__stub.ListServingSites(request)
            return [ServingSite(item, self) for item in response.servingsites.items]
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False





import grpc
from github.com.metaprov.modelaapi.pkg.apis.data.v1alpha1.generated_pb2 import FeatureHistogram as MDFeatureHistogram
from github.com.metaprov.modelaapi.services.featurehistogram.v1.featurehistogram_pb2_grpc import FeatureHistogramServiceStub
from github.com.metaprov.modelaapi.services.featurehistogram.v1.featurehistogram_pb2 import CreateFeatureHistogramRequest, \
    UpdateFeatureHistogramRequest, \
    DeleteFeatureHistogramRequest, GetFeatureHistogramRequest, ListFeatureHistogramsRequest

from modela.Resource import Resource
from modela.ModelaException import ModelaException
from typing import List, Union


class FeatureHistogram(Resource):
    def __init__(self, item: MDFeatureHistogram = MDFeatureHistogram(), client=None, namespace="", name="", version=Resource.DefaultVersion):
        super().__init__(item, client, namespace=namespace, name=name, version=version)


class FeatureHistogramClient:
    def __init__(self, stub, modela):
        self.modela = modela
        self.__stub: FeatureHistogramServiceStub = stub

    def create(self, featurehistogram: FeatureHistogram) -> bool:
        request = CreateFeatureHistogramRequest()
        request.featurehistogram.CopyFrom(featurehistogram.raw_message)
        try:
            response = self.__stub.CreateFeatureHistogram(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def update(self, featurehistogram: FeatureHistogram) -> bool:
        request = UpdateFeatureHistogramRequest()
        request.featurehistogram.CopyFrom(featurehistogram.raw_message)
        try:
            self.__stub.UpdateFeatureHistogram(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def get(self, namespace: str, name: str) -> Union[FeatureHistogram, bool]:
        request = GetFeatureHistogramRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.GetFeatureHistogram(request)
            return FeatureHistogram(response.featurehistogram, self)
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def delete(self, namespace: str, name: str) -> bool:
        request = DeleteFeatureHistogramRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.DeleteFeatureHistogram(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def list(self, namespace: str) -> Union[List[FeatureHistogram], bool]:
        request = ListFeatureHistogramsRequest()
        request.namespace = namespace
        try:
            response = self.__stub.ListFeatureHistograms(request)
            return [FeatureHistogram(item, self) for item in response.featurehistograms.items]
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False





# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['modela',
 'modela.data',
 'modela.inference',
 'modela.infra',
 'modela.team',
 'modela.training']

package_data = \
{'': ['*']}

install_requires = \
['grpcio>=1.44.0,<2.0.0',
 'kubernetes>=21.7.0,<22.0.0',
 'modelaapi==0.4.567',
 'protobuf>=3.19.3,<4.0.0']

setup_kwargs = {
    'name': 'modela',
    'version': '0.4.567',
    'description': 'python client sdk for the modela auto ml system.',
    'long_description': '## What is Modela?\n\nModela is a Kubernetes-based automatic machine learning platform that enables you with an expansive API for machine\nlearning. The system implements Kubernetes custom resources for declaratively managing data, training, inference, and \ninfrastructure objects. The Modela control plane manages these resources to provide fully distributed and automated machine learning. \n\n\nFor a complete documentation of the API, check out the [documentation](https://www.modela.ai/docs/docs).\n\n## Install Modela\n\nModela can be installed through the Modela CLI, or through Helm Charts. For a complete installation procedure, please \nrefer to the installation procedure on the [documentation](https://www.modela.ai/docs/docs/install).\n\n## Using the SDK\n\nThe Modela API is implemented by [Modela](https://modela.ai), which can be installed on \nany Kubernetes cluster for free. You can access the API by connecting to the API gateway service hosted by your installation.\n\n',
    'author': 'liam',
    'author_email': 'liam@metaprov.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/metaprov/modela-python-sdk',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import grpc


class ModelaException(Exception):
    """
    Error handling for gRPC and application-level errors is done with exceptions. ModelaException provides a base class
    for all exceptions raised by the Modela SDK.
    """

    def __init__(self, code: int, error: str, details: str, debug_string: str):
        super().__init__()
        self._code = code
        self._error = error
        self._details = details
        self._debug_string = debug_string

    @property
    def code(self):
        """
        The error code returned by the gRPC interface
        """
        return self._code

    @property
    def error(self):
        """
        The error code name returned by the gRPC interface
        """
        return self._error

    @property
    def details(self):
        """
        The error details returned by the Modela API gateway
        """
        return self._details

    @property
    def debug_string(self):
        """
        The error debug string returned by gRPC
        """
        return self._debug_string

    @staticmethod
    def process_error(err: grpc.RpcError):
        if err.code() == grpc.StatusCode.ALREADY_EXISTS.value:
            raise ResourceExistsException(err.code().value, err.code().name, err.details(), err.debug_error_string())
        if err.code() == grpc.StatusCode.NOT_FOUND:
            raise ResourceNotFoundException(err.code().value, err.code().name, err.details(), err.debug_error_string())
        elif err.code() == grpc.StatusCode.UNAUTHENTICATED:
            raise UnauthenticatedException(err.code().value, err.code().name, err.details(), err.debug_error_string())
        elif err.code() == grpc.StatusCode.PERMISSION_DENIED:
            raise PermissionDeniedException(err.code().value, err.code().name, err.details(), err.debug_error_string())
        elif err.code() == grpc.StatusCode.FAILED_PRECONDITION:
            raise InvalidResourceException(err.code().value, err.code().name, err.details(), err.debug_error_string())
        else:
            raise GrpcErrorException(err.code().value, err.code().name, err.details(), err.debug_error_string())

    def __str__(self):
        return "{details} ({error}, {code})".format(details=self.details, error=self.error, code=self.code)


class ResourceNotFoundException(ModelaException):
    """
    Exception raised in the case of a resource not being found in a namespace with a given name.
    """


class ResourceExistsException(ModelaException):
    """
    Exception raised in the case of a resource being created when it already exists.
    """


class UnauthenticatedException(ModelaException):
    """
    Exception raised in the case of the user being unauthenticated.
    """


class PermissionDeniedException(ModelaException):
    """
    Exception raised in the case of unauthorized access to a certain type of resource action.
    """


class InvalidResourceException(ModelaException):
    """
    Exception raised in the case of the application recognizing a resource update or create request to be invalid.
    """


class GrpcErrorException(ModelaException):
    """
    Exception raised in the case of any miscellaneous gRPC error
    """

